; Emacs support for LAML midi processing.
; Load this file in your emacs init file:  .emacs

(provide 'laml-midi-support)

; ---------------------------------------------------------------------------------------------------------------
; Configuration part:

(defvar midi-player-exc "c:/program files/Winamp/winamp.exe" "The full path to your midi player - preferable winamp")

(defvar midi-illustrator-exc "c:/Program Files/Rallentando Software/MidiIllustrator 2/MI.exe" "The full path to a midi illustrator - score presentation program.")


; ---------------------------------------------------------------------------------------------------------------

(defvar path-to-midi-error-file (concat laml-dir "styles/xml-in-laml/midi/" "midi/error.mid")
        "A path to a midi error file, which is played if no mirror file can be generated")  


(defvar laml-temp-file-path (concat laml-dir "temp/") "The LAML temp directory. Assumed to be writable.")


; Midi Laml Mode stuff and key/menu bidings.
; The mode is backed up by the file extension "midl".

(make-variable-buffer-local 'local-midl-mode-info)

(define-derived-mode 
	midi-laml-mode laml-mode "MidiLaml" 
  "The MidiLaml mode is a specialization of Laml mode which supports programmed MIDI solutions
in a particular MIDI XML-in-LAML language.
The LAML MIDI processor is a part of the LAML distribution.
LAML means 'Lisp Abstracted Markup Language'. LAML makes HTML and XML languages available as 
Scheme functions."

;   actions-here
  (make-variable-buffer-local 'attribute-to-be-modified)
  (make-variable-buffer-local 'play-after-navigate)
  (make-variable-buffer-local 'channels-to-play)   

  (show-midl-mode-status-string)

  (setq mode-line-format 
     '("%e"
       #("-" 0 1
         (help-echo "mouse-1: select (drag to resize), mouse-2 = C-x 1, mouse-3 = C-x 0"))
       mode-line-mule-info mode-line-modified mode-line-frame-identification mode-line-buffer-identification
       #("   " 0 3
         (help-echo "mouse-1: select (drag to resize), mouse-2 = C-x 1, mouse-3 = C-x 0"))
       mode-line-position
       (vc-mode vc-mode)
       #("  " 0 2
         (help-echo "mouse-1: select (drag to resize), mouse-2 = C-x 1, mouse-3 = C-x 0"))
       mode-line-modes
       (which-func-mode
        ("" which-func-format
         #("--" 0 2
           (help-echo "mouse-1: select (drag to resize), mouse-2 = C-x 1, mouse-3 = C-x 0"))))
       (:eval (midl-mode-status-string))
       (global-mode-string
        (#("--" 0 2
           (help-echo "mouse-1: select (drag to resize), mouse-2 = C-x 1, mouse-3 = C-x 0"))
         global-mode-string))
       #("-%-" 0 3
         (help-echo "mouse-1: select (drag to resize), mouse-2 = C-x 1, mouse-3 = C-x 0"))))

)

(setq auto-mode-alist 
      (cons '("\\.midl" . midi-laml-mode)
	    auto-mode-alist))

(defun size-of-current-buffer ()
   (point-max))

; Midi LAML keymap:
(defvar midi-laml-mode-map (make-sparse-keymap "MidiLaml"))



(laml-define-key midi-laml-mode-map [menu-bar midi-laml-stack]
    (cons "MIDL Stack" (make-sparse-keymap "MIDL Stack")))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml-comments]
    (cons "MIDL Comments" (make-sparse-keymap "MIDL Comments")))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml]
    (cons "MIDL" (make-sparse-keymap "MIDL")))

; (laml-define-key midi-laml-mode-map "\C-i" 'goto-next-input-field)

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml help-on-function-keys]
      '("Help on Function Keys" . (lambda () (interactive) (help-from-text-file "function-keys.txt"))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml help-on-markers]
      '("Help on Markers" . (lambda () (interactive) (help-from-text-file "marker-channel.txt"))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml midi-laml-manual]
      '("Browse MIDI LAML Manual" . (lambda () (interactive) (open-midi-html-documentation "midi-mirror.html"))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml midi-laml-function-manual]
      '("Browse MIDI Function Library Manual" . (lambda () (interactive) (open-midi-html-documentation "midi-laml-processing-lib.html"))))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml menu-midi-laml-sep-11]
    '("----"))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml insert-template]
    (cons "Insert template" (make-sparse-keymap "Insert template")))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml menu-midi-laml-sep-10]
    '("----"))


(laml-define-key  midi-laml-mode-map [menu-bar midi-laml insert-template tyros-multi-pad]
      '("Tyros multipad" . 
          (lambda () (interactive)
            (insert-file-contents (concat laml-dir "styles/xml-in-laml/midi/" "templates/" "tyros-multipad" "." "tpl")))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml insert-template tyros-style]
      '("Tyros style - absTime" . 
          (lambda () (interactive)
            (insert-file-contents (concat laml-dir "styles/xml-in-laml/midi/" "templates/" "tyros-style-abs-time" "." "tpl")))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml insert-template simple-format-1-16]
      '("Format 1 midi file - 16 tracks" . 
          (lambda () (interactive)
            (insert-file-contents (concat laml-dir "styles/xml-in-laml/midi/" "templates/" "simple-midi-format-1-16" "." "tpl")))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml insert-template simple-format-1-2]
      '("Format 1 midi file - 2 tracks" . 
          (lambda () (interactive)
            (insert-file-contents (concat laml-dir "styles/xml-in-laml/midi/" "templates/" "simple-midi-format-1-2" "." "tpl")))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml insert-template simple-format-0]
      '("Format 0 midi file" . 
          (lambda () (interactive)
            (insert-file-contents (concat laml-dir "styles/xml-in-laml/midi/" "templates/" "simple-midi-format-0" "." "tpl")))))


(laml-define-key  midi-laml-mode-map [menu-bar midi-laml toggle-syntax-highlighting]
      '("Toggle syntax highlighting" . toggle-syntax-highlighting))


(laml-define-key  midi-laml-mode-map [menu-bar midi-laml format-attributes-of-multi-pad]
      '("Format attributes of multipad form" . format-attributes-of-multi-pad))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml menu-midi-laml-sep-9]
    '("----"))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml make-marker-region-browser]
      '("Marker browser of selected track" . make-marker-region-browser))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml menu-midi-laml-sep-12]
    '("----"))


(laml-define-key  midi-laml-mode-map [menu-bar midi-laml display-length-of-midi-list]
      '("Display length of selection" . display-length-of-midi-list))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml menu-midi-laml-sep-8]
    '("----"))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml goto-bar]
      '("Goto bar..." . goto-bar))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml menu-midi-laml-sep-7]
    '("----"))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml outdent-channel]
      '("Outdent channel of track..." . outdent-channel))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml outdent-multiple-channels]
      '("Outdent multiple channels of track..." . outdent-multiple-channels))

; (laml-define-key midi-laml-mode-map [menu-bar midi-laml menu-midi-laml-sep-6]
;    '("----"))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments remove-bars-of-midi-track-liberal]
      '("Remove bar comments from current track" . remove-bars-of-midi-track-liberal))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments draw-bars-of-midi-track-liberal]
      '("Add bar comments to current track" . draw-bars-of-midi-track-liberal))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml-comments menu-midi-laml-sep-1]
   '("----"))

; (laml-define-key midi-laml-mode-map [menu-bar midi-laml score-comment]
;    (cons "Score Comments" (make-sparse-keymap "Score Comments")))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml-comments set-c]
    (cons "Set lower C" (make-sparse-keymap "Set lower C for score comments")))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments report-lower-c]
      '("What is lower C" . report-lower-c))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments set-c set-c-to-c7]
      '("Set lower C to C7 (96)" . (lambda () (interactive) (set-c-value 96))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments set-c set-c-to-c6]
      '("Set lower C to C6 (84)" . (lambda () (interactive) (set-c-value 84))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments set-c set-c-to-c5]
      '("Set lower C to C5 (72)" . (lambda () (interactive) (set-c-value 72))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments set-c set-c-to-c4]
      '("Set lower C to C4 (60)" . (lambda () (interactive) (set-c-value 60))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments set-c set-c-to-c3]
      '("Set lower C to C3 (48)" . (lambda () (interactive) (set-c-value 48))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments set-c set-c-to-c2]
      '("Set lower C to C2 (36)" . (lambda () (interactive) (set-c-value 36))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments set-c set-c-to-c1]
      '("Set lower C to C1 (24)" . (lambda () (interactive) (set-c-value 24))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments set-c set-c-to-c0]
      '("Set lower C to C0 (12)" . (lambda () (interactive) (set-c-value 12))))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments remove-score-comments-of-bars]
      '("Remove score comments of bars..." . remove-score-comments-of-bars))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments remove-score-comments-of-region]
      '("Remove score comments of region" . remove-score-comments-of-region))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments remove-score-comments-of-form]
      '("Remove score comments from form" . remove-score-comments-of-form))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments add-score-comments-of-bar-channel]
      '("Add score comments to channel of bars..." . add-score-comments-of-bar-channel))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments add-score-comments-of-region-channel]
      '("Add score comments to channel of region..." . add-score-comments-of-region-channel))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments add-score-comments-of-form-channel]
      '("Add score comments to channel of form..." . add-score-comments-of-form-channel))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments add-score-comments-to-bars]
      '("Add score comments to bars..." . add-score-comments-to-bars))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments add-score-comments-of-region]
      '("Add score comments to region" . add-score-comments-of-region))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments add-score-comments-of-form]
      '("Add score comments to form" . add-score-comments-of-form))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml-comments menu-midi-laml-sep-2]
    '("----"))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments add-comment-to-midl-file]
      '("Add midl file comment..." . add-comment-to-midl-file))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-comments display-comment-of-midl-file]
      '("Show midl file comment" . display-comment-of-midl-file))




; (laml-define-key midi-laml-mode-map [menu-bar midi-laml baseline]
;    (cons "Baseline" (make-sparse-keymap "Baseline")))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack baseline-present-info]
      '("Info about baselines" . baseline-present-info))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack push-baseline-on-midl-stack]
      '("Push baseline file on midl stack..." . push-baseline-on-midl-stack))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack create-baseline-of-current-midl-source-file]
      '("Create baseline file..." . create-baseline-of-current-midl-source-file))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml-stack menu-midi-laml-sep]
    '("----"))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack help-on-midl-stack]
      '("Help on Midl Stack" . (lambda () (interactive) (help-from-text-file "midl-stack.txt"))))

; (laml-define-key midi-laml-mode-map [menu-bar midi-laml midl-stack]
;    (cons "Midl Stack" (make-sparse-keymap "Midl Stack")))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack midl-source-file-stack-info]
      '("Info about midl stack" . midl-source-file-stack-info))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml-stack menu-midi-laml-sep-14]
    '("----"))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack reset-midl-ghost-stack]
      '("Get rid of ghost elements" . reset-midl-ghost-stack))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack reset-midl-source-file-stack]
      '("Reset midl stack..." . reset-midl-source-file-stack))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml-stack menu-midi-laml-sep-15]
    '("----"))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack undo-pop-midl-stack]
      '("Undo pop" . undo-pop-midl-stack))

; Done implicitly, when necessary:
; (laml-define-key  midi-laml-mode-map [menu-bar midi-laml recover-midl-source-file-stack]
;       '("Recover midl stack" . recover-midl-source-file-stack))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack merge-two-top-stack-entries]
      '("Pop splitted entries, merge and push result" . merge-two-top-stack-entries))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack pop-and-reestablish-midl-file]
      '("Pop and reestablish source" . pop-and-reestablish-midl-file))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml-stack menu-midi-laml-sep-13]
    '("----"))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack push-latest-temporary-selection-on-midl-stack]
      '("Push latest temp selection" . push-latest-temporary-selection-on-midl-stack))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack push-external-midl-file-on-midl-stack]
      '("Push external midl file" . push-external-midl-file-on-midl-stack))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack replicate-current-buffer-and-undo-all-pops]
      '("Replicate current buffer - undo all pops" . replicate-current-buffer-and-undo-all-pops))


(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack split-and-replicate-midl-file-by-channel]
      '("Process, save, and push splits by channel..." . split-and-replicate-midl-file-by-channel))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack process-laml-file-save-and-push-with-counter-transposition]
      '("Process, save and push source - counter transposition" . process-laml-file-save-and-push-with-counter-transposition))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack process-laml-file-save-and-push-with-new-processing-mode]
      '("Process, save and push source - new top mode" . process-laml-file-save-and-push-with-new-processing-mode))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml-stack process-laml-file-save-and-push]
      '("Process, save and push source" . process-laml-file-save-and-push))

(laml-define-key midi-laml-mode-map [menu-bar midi-laml menu-midi-laml-sep-2]
    '("----"))

; (laml-define-key  midi-laml-mode-map [menu-bar midi-laml next-similar-midi-step]
;      '("Next similar midi step" . next-similar-midi-step))

; (laml-define-key  midi-laml-mode-map [menu-bar midi-laml midi-step]
;      '("Midi step" . midi-step))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-selected-laml-file-and-play-again]
      '("Process and Play selection again" . process-selected-laml-file-and-play-again))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-selected-laml-file-certain-channels-and-illustrate]
      '("Process and Illustrate selection of channels..." . process-selected-laml-file-certain-channels-and-illustrate))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-bars-laml-file-certain-channels-and-illustrate]
      '("Process and Illustrate bars of channels..." . process-bars-laml-file-certain-channels-and-illustrate))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-selected-track-and-illustrate]
      '("Process and Illustrate track" . process-selected-track-and-illustrate))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-selected-laml-file-certain-channels-and-play]
      '("Process and Play selection of channels..." . process-selected-laml-file-certain-channels-and-play))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-bars-laml-file-certain-channels-and-play]
      '("Process and Play bars of channels..." . process-bars-laml-file-certain-channels-and-play))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-selected-laml-file-and-play]
      '("Process and Play selection" . process-selected-laml-file-and-play))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-selected-bars-laml-file-and-play]
      '("Process and Play bars..." . process-selected-bars-laml-file-and-play))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-selected-track-and-play]
      '("Process and Play track" . process-selected-track-and-play))


(laml-define-key midi-laml-mode-map [menu-bar midi-laml menu-midi-laml-sep-1]
    '("----"))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-laml-file-and-save]
      '("Process and Save ..." . process-laml-file-and-save))

(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-laml-file-and-illustrate]
      '("Process and Illustrate" . process-laml-file-and-illustrate))


(laml-define-key  midi-laml-mode-map [menu-bar midi-laml process-laml-file-and-play]
      '("Process and Play" . process-laml-file-and-play))

(laml-define-key midi-laml-mode-map (list "\C-o" "\C-c\C-o") 'process-laml-file-and-play)
(laml-define-key midi-laml-mode-map (list "\C-c\C-c" "\C-c\C-c") 'process-selected-laml-file-and-play)
(laml-define-key midi-laml-mode-map (list "\C-c\C-b" "\C-c\C-b") 'process-selected-bars-laml-file-and-play)
(laml-define-key midi-laml-mode-map (list "\C-c\C-t" "\C-c\C-t") 'process-selected-track-and-play)
(laml-define-key midi-laml-mode-map (list "\C-c\C-k" "\C-c\C-k") 'process-selected-laml-file-and-play-again)
(laml-define-key midi-laml-mode-map (list "\C-c\C-m" "\C-c\C-m") 'midi-step)

(laml-define-key midi-laml-mode-map (list "\C-c\C-u" "\C-c\C-u") 'next-similar-midi-step)
(laml-define-key midi-laml-mode-map (list "\C-c\C-s" "\C-c\C-s") 'process-laml-file-and-save)
(laml-define-key midi-laml-mode-map (list "\C-c\C-l" "\C-c\C-l") 'process-laml-file-and-illustrate)

(laml-define-key midi-laml-mode-map (list "\C-c\C-p" "\C-c\C-p") 'process-laml-file-save-and-push)
(laml-define-key midi-laml-mode-map (list "\C-cs" "\C-cs") 'split-and-replicate-midl-file-by-channel)
(laml-define-key midi-laml-mode-map (list "\C-c\C-o" "\C-c\C-o") 'pop-and-reestablish-midl-file)
(laml-define-key midi-laml-mode-map (list "\C-c\C-y" "\C-c\C-y") 'undo-pop-midl-stack)
(laml-define-key midi-laml-mode-map (list "\C-c\C-i" "\C-c\C-i") 'midl-source-file-stack-info)
(laml-define-key midi-laml-mode-map (list "\C-ct" "\C-ct")       'push-latest-temporary-selection-on-midl-stack)
(laml-define-key midi-laml-mode-map (list "\C-cP" "\C-cP") 'push-external-midl-file-on-midl-stack)
(laml-define-key midi-laml-mode-map (list "\C-cR" "\C-cR") 'replicate-current-buffer-and-undo-all-pops)
(laml-define-key midi-laml-mode-map (list "\C-cc" "\C-cc") 'display-comment-of-midl-file)
(laml-define-key midi-laml-mode-map (list "\C-c\C-a" "\C-c\C-a") 'add-comment-to-midl-file)

(laml-define-key midi-laml-mode-map (list "\C-cb" "\C-cb") 'create-baseline-of-current-midl-source-file)
(laml-define-key midi-laml-mode-map (list "\C-ci" "\C-ci") 'baseline-present-info)

(laml-define-key midi-laml-mode-map (list "\C-cm" "\C-cm") 'make-marker-region-browser)
(laml-define-key midi-laml-mode-map (list "\C-cl" "\C-cl") 'display-length-of-midi-list)

; Function keys: 

; F5-F8: State and toggles: 
(define-key  midi-laml-mode-map [f7] 'set-channels-to-play)
(define-key  midi-laml-mode-map [f6] 'toggle-play-after-navigate)
(define-key  midi-laml-mode-map [f5] 'toggle-attribute-to-be-modified)

; ---
; (define-key  midi-laml-mode-map [f5] 'previous-noteon-and-play)
; (define-key  midi-laml-mode-map [f6] 'play-noteon)
; (define-key  midi-laml-mode-map [f7] 'next-noteon-and-play)
; 
; (define-key  midi-laml-mode-map [f9] 'goto-previous-bar-and-process-current-bar-and-play)
; (define-key  midi-laml-mode-map [f10] 'process-current-bar-and-play)
; (define-key  midi-laml-mode-map [f11] 'goto-next-bar-and-process-current-bar-and-play)
; ---

(define-key  midi-laml-mode-map [f9] 'play-noteon)
(define-key  midi-laml-mode-map [f10] 'process-current-bar-and-play)
(define-key  midi-laml-mode-map [f11] 'process-from-current-bar-to-end-and-play)
(define-key  midi-laml-mode-map [f12] 'process-selected-laml-file-and-play-again)


; Arrows:
(define-key  midi-laml-mode-map [up] 'previous-noteon-and-play)
(define-key  midi-laml-mode-map [kp-space] 'play-noteon)  ; Key pad 5
(define-key  midi-laml-mode-map [down] 'next-noteon-and-play)

(define-key  midi-laml-mode-map [C-up] 'goto-previous-bar-and-process-current-bar-and-play)
(define-key  midi-laml-mode-map [C-kp-up] 'goto-previous-bar-and-process-current-bar-and-play)
(define-key  midi-laml-mode-map [M-C-up] 'goto-previous-marked-section-and-process-it-and-play)
(define-key  midi-laml-mode-map [M-C-kp-up] 'goto-previous-marked-section-and-process-it-and-play)

(define-key  midi-laml-mode-map [C-kp-space] 'process-current-bar-and-play)    ; Key pad 5
(define-key  midi-laml-mode-map [M-C-kp-space] 'process-current-marked-region-and-play)    ; Key pad 5

(define-key  midi-laml-mode-map [C-down] 'goto-next-bar-and-process-current-bar-and-play)
(define-key  midi-laml-mode-map [C-kp-down] 'goto-next-bar-and-process-current-bar-and-play)
(define-key  midi-laml-mode-map [M-C-down] 'goto-next-marked-section-and-process-it-and-play)
(define-key  midi-laml-mode-map [M-C-kp-down] 'goto-next-marked-section-and-process-it-and-play)

(define-key  midi-laml-mode-map [left] '(lambda () (interactive) (do-subtract-from-attribute-of-current-form 1)))
(define-key  midi-laml-mode-map [C-left] '(lambda () (interactive) (do-subtract-from-attribute-of-current-form 2)))
(define-key  midi-laml-mode-map [C-kp-left] '(lambda () (interactive) (do-subtract-from-attribute-of-current-form 2)))
(define-key  midi-laml-mode-map [C-M-left] '(lambda () (interactive) (do-subtract-from-attribute-of-current-form 3)))
(define-key  midi-laml-mode-map [C-M-kp-left] '(lambda () (interactive) (do-subtract-from-attribute-of-current-form 3)))

(define-key  midi-laml-mode-map [right] '(lambda () (interactive) (do-add-to-attribute-of-current-form 1)))
(define-key  midi-laml-mode-map [C-right] '(lambda () (interactive) (do-add-to-attribute-of-current-form 2)))
(define-key  midi-laml-mode-map [C-kp-right] '(lambda () (interactive) (do-add-to-attribute-of-current-form 2)))
(define-key  midi-laml-mode-map [C-M-right] '(lambda () (interactive) (do-add-to-attribute-of-current-form 3)))
(define-key  midi-laml-mode-map [C-M-kp-right] '(lambda () (interactive) (do-add-to-attribute-of-current-form 3)))

; -----------------------------------------------------------------------------

; Emacs tools menu contributions:

(laml-define-key global-map [menu-bar tools midi-laml-menu]
    (cons "Midi Laml" (make-sparse-keymap "Midi Laml")))

(laml-define-key  global-map [menu-bar tools midi-laml-menu new-multi-pad]
      '("Make tyros multipad..." . (lambda () (interactive) (make-new-midl-file "tyros-multipad"))))

(laml-define-key  global-map [menu-bar tools midi-laml-menu new-abs-time-style]
      '("Make tyros absTime style..." . (lambda () (interactive) (make-new-midl-file "tyros-style-abs-time"))))

(laml-define-key  global-map [menu-bar tools midi-laml-menu new-format-1-16]
      '("Make format 1 midi file - 16 tracks..." . (lambda () (interactive) (make-new-midl-file "simple-midi-format-1-16"))))

(laml-define-key  global-map [menu-bar tools midi-laml-menu new-format-1-2]
      '("Make format 1 midi file - 2 tracks..." . (lambda () (interactive) (make-new-midl-file "simple-midi-format-2"))))

(laml-define-key  global-map [menu-bar tools midi-laml-menu new-format-0-midi]
      '("Make format 0 midi file..." . (lambda () (interactive) (make-new-midl-file "simple-midi-format-0"))))


(defun buffer-reset-undo ()
  "Discard undo information in the current buffer"
  (buffer-disable-undo)
  (buffer-enable-undo))



(defun toggle-syntax-highlighting ()
  (interactive)
  (font-lock-mode nil)
)

(defun make-new-midl-file (template-name)
 (let ((use-dialog-box nil))  ; Prevents Windows Explorer Dialog - Per dynamic binding!!!
  (let* ((dir0 (call-interactively (function (lambda (dir) (interactive "DMake a midi laml file in which directory: ") dir))))
         (name (call-interactively (function (lambda (name) (interactive "sName of midi laml file (without extension): ") name))))
         (dir (ensure-trailing-slash (expand-file-name dir0)))
         (new-buffer (create-file-buffer (concat dir name "." "midl"))))
    (set-buffer new-buffer)
    (write-file (concat dir name "." "midl"))
    (insert-file-contents (concat laml-dir "styles/xml-in-laml/midi/" "templates/" template-name "." "tpl"))
    (midi-laml-mode)
    (show-buffer (selected-window) new-buffer)
    new-buffer)))



; ---------------------------------------------------------------------------------------------------------------


(defun midi-to-laml (processing-mode counter-transposition) 
  "From dired, process the selected midi file to the corresponding LAML file. Create a midi-temp directory for
temporary files."
 (interactive (list 
                (completing-read "Processing mode: " '(("absTime" . nil) ("absTime-on-off" . nil) ("deltaTime" . nil) ("raw" . nil)) nil nil "absTime")
                (completing-read "Counter transposition: " '() nil nil "0")))
 
 ; Ensure that the midi-temp directory exists
 (if (not (file-exists-p (concat (current-directory) "midi-temp/")))   ; directory existence
     (make-directory (concat (current-directory) "midi-temp/")))
 
 (let ((selected-files (dired-get-marked-files nil)))
   (if (and (= 1 (length selected-files)) (file-regular-p (car selected-files)))
       (let* ((full-path-midi-file (car selected-files))
              (ext (file-name-extension full-path-midi-file)))
         (if (midi-extension-p ext)
             (let ((track-separation (ask-user "Should the midi file be track separated [yes/no]?" "no")))
               (midi-file-to-laml-file processing-mode counter-transposition track-separation full-path-midi-file))
             (midi-file-to-laml-file processing-mode counter-transposition nil full-path-midi-file))
         (other-window 1)
       )
     (progn
       (beep)
       (message "Cannot process. You should only select a single midi file.")))))

(defun midi-extension-p (x)
  (or (equal (downcase x) "mid") (equal (downcase x) "midi")))

; Asynchronous midi file to laml file processing via laml temp dir.
(defun midi-file-to-laml-file (processing-mode counter-transposition track-separation full-path-midi-file) 
  (let ((laml-temp-processing-dir (concat laml-dir "temp/"))) 
    (make-a-file-from-laml-template 
       "temp-processing-program.laml"
       laml-temp-processing-dir
       "midi-to-laml-processor"
       'laml-mode
       (list
        (list "FULL-PATH-TO-MIDI-FILE" full-path-midi-file)
        (list "THE-PROCESSING-MODE" processing-mode)
        (list "THE-COUNTER-TRANSPOSITION" counter-transposition)
        (list "THE-TRACK-SEPARATION" (if track-separation "true" "false"))
        )
       (concat laml-dir "styles/xml-in-laml/midi/" "templates/"))  
    (laml-process-file  (concat laml-temp-processing-dir "temp-processing-program.laml"))))

; Synchronous version with more parameters
(defun sync-midi-file-to-laml-file (processing-mode counter-transposition track-separation full-path-midi-file full-path-laml-file) 
  (let ((laml-temp-processing-dir (concat laml-dir "temp/"))) 
    (make-a-file-from-laml-template 
       "temp-processing-program.laml"
       laml-temp-processing-dir
       "sync-midi-to-laml-processor"
       'laml-mode
       (list
        (list "FULL-PATH-TO-MIDI-FILE" full-path-midi-file)
        (list "FULL-PATH-TO-LAML-FILE" full-path-laml-file)
        (list "THE-PROCESSING-MODE" processing-mode)
        (list "THE-COUNTER-TRANSPOSITION" counter-transposition)
        (list "THE-TRACK-SEPARATION" (if track-separation "true" "false"))
        )
       (concat laml-dir "styles/xml-in-laml/midi/" "templates/"))  
    (laml-sync-process-file (concat laml-temp-processing-dir "temp-processing-program.laml") 'silent)))


; The function midi-to-laml is bound to a dired menu entry in the function laml-enable-dired-mode of
; the file emacs-support/laml-key-menu-bindings.el.
; The file laml-key-menu-bindings.el is part of the LAML distribution.

; ---------------------------------------------------------------------------------------------------------------

(defun outdent-channel (n)
 "Outdent all midi messages that belong to channel n"
 (interactive "nOutdent channel: ")
 (message (concat "Outdenting channel " (number-to-string n) "..."))
 (save-excursion
  (if (not (looking-at "(MidiTrack"))
      (backward-up-list 1))
  (let ((limit (point))
        (done nil))
    (forward-sexp 1) (backward-char 1)
    
    (while (not done)
      (if (re-search-backward 
           (regular-expression 
            `(concat "'channel" (one-or-more " ") (concat "\"") ,(int-to-string n) (concat "\"")))
           limit t)
          (progn
            (beginning-of-line 1)
            (indent-amount -2)
          )
          (setq done t)))))
  (message "DONE"))

(defun outdent-multiple-channels (str-int-lst)
  (interactive "sOutent which channels (numbers separated by space): ")
  (message (concat "Outdenting channels " str-int-lst "..."))
  (save-excursion
    (let ((n-list (car (read-from-string (concat "(" str-int-lst ")")))))
      (mapcar (function outdent-channel) n-list)))
  (message "DONE"))
    


(defun indent-amount (a)
  (if (> a 0)
      (insert (make-string a 32))
      (let ((count 0))
        (while (and (< count (abs a)) (looking-at " "))
           (delete-char 1)
           (setq count (+ count 1))))))



; ---------------------------------------------------------------------------------------------------------------

(defun format-attributes-of-multi-pad ()
  "Reformat the attributes of a TyrosMulitiPad form."
  (interactive)

  (save-excursion
    (goto-char (point-min))
    (let ((res (search-forward "'name-1" nil t)))
       (if res
           (progn
             (beginning-of-line 1)

             (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)
             (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line)
             (forward-char 1)

             (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)
             (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line)
             (forward-char 1)

             (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)
             (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line)
             (forward-char 1)

             (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)
             (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line 1) (delete-char 1)   (end-of-line)
             (forward-char 1))
           (progn
             (beep)
             (message "You can only use this command on a buffer with multipad content"))))))

; ---------------------------------------------------------------------------------------------------------------
; OVERLAY STUFF. Not used.

(defvar overlayable-forms nil)

(setq overlayable-forms
  (regular-expression
    '(concat "(" (one-or-more " ")
       (or "ProgramChange" "ControlChange" "NoteOn" "NoteOff" "SysEx" "Meta")   ; more
    )))

(defun prepare-overlaying ()
 (interactive)
 (remove-overlays-from-current-buffer)
 (setq buffer-invisibility-spec nil)
 (let ((count 0))
 (save-excursion
  (goto-char (point-min))
  (let ((done nil))
    (while (not done)
      (if (re-search-forward overlayable-forms nil t)
        (let* ((kind (type-of-overlay-under (point)))
               (ext (extension-of-overlay-under (point)))
               (overlay (make-overlay (car ext) (cdr ext) (current-buffer))))
          (overlay-put overlay 'invisible kind)
          (overlay-put overlay 'intangible nil)   ; ???
          (setq count (+ count 1))
          (message (number-to-string count))
        )
        (setq done t)))))))

(defun remove-overlays-from-current-buffer()
  (let ((overlay-list (overlays-in (point-min) (point-max))))
    (mapcar (function (lambda (o) (delete-overlay o))) overlay-list)))

; Return the type of the overlay of the form under point. 
; Point is assumed to be inside the form (a XML-in-LAML midi message).
(defun type-of-overlay-under (p)
  (save-excursion 
    (backward-up-list 1)
    (forward-char)
    (let ((name-of-form (string-to-symbol (prog-name-under-point))))
      (backward-up-list 1)
      (let ((p0 (point)))
        (forward-sexp 1)
        (let ((p1 (point)))
          (backward-sexp 1)
          (let ((channel-p (search-forward "'channel" p1 t)))
            (if channel-p
                (int-value-of-string-ahead)
              name-of-form)))))))

(defun int-value-of-string-ahead ()
  (progn
    (search-forward "\"")
    (let ((p2 (point)))
      (search-forward "\"")
      (string-to-int (buffer-substring p2 (- (point) 1))))))

; Return a cons pair of start and end position of the overlay of the form under point. 
; Point is assumed to be inside the form (a XML-in-LAML midi message).
(defun extension-of-overlay-under (p)
  (save-excursion 
    (backward-up-list 1)
    (let ((p0 (point)))
      (forward-sexp 1)
      (end-of-line 1) 
      (while (looking-at-chars (point) white-space-char-list) (forward-char 1))
      (cons p0 (point)))))

; delete:
; (defvar the-visibile-channel t)

(defun show-only-channel (n)
  "Show only channel n"
  (interactive "nShow only channel: ")
  (show-all-channels)
  (add-to-invisibility-spec 'SysEx)   (add-to-invisibility-spec 'Meta)
  (setq the-visible-channel n)
  (mapcar 
   (lambda (ch) 
     (if (not (= n ch))
       (add-to-invisibility-spec ch)))
   (number-interval 1 16)))

(defun show-all-channels ()
  (interactive)
  (setq buffer-invisibility-spec nil))

(defun next-visible-line ()
  (let* ((overlays (overlays-at (point))))
    (if (not (null overlays))
        (let ((overlay-props (overlay-get (car overlays) 'invisible)))
          (next-line 1)
          (if (not (eq overlay-props the-visible-channel))
              (next-visible-line)))
        (next-line 1))))


; ---------------------------------------------------------------------------------------------------------------
; Midi playing

(defun play-midi-file (midi-file-path)
  "Play midi file in midi-file-path with a midi player selected via the variable midi-player-exc"
  (interactive "f:Play MIDI file: ")
  (start-process "MIDI play" nil midi-player-exc midi-file-path))


(defun illustrate-midi-file (midi-file-path)
  "Illustrate midi file in midi-file-path with a midi illustrator selected via the variable midi-illustrator-exc"
  (interactive "f:Illustrate MIDI file: ")
  (start-process "MIDI illustrate" nil midi-illustrator-exc midi-file-path))

(defun play-midi-file-from-dired ()
  (interactive)
  (let ((selected-files (dired-get-marked-files nil)))
    (if (and (= 1 (length selected-files)) (file-regular-p (car selected-files)))
      (let* ((full-path-midi-file (car selected-files)))
         (play-midi-file full-path-midi-file))
      (progn
        (beep)
        (message "Cannot play. You should only select a single midi file.")))))

(defun illustrate-midi-file-from-dired ()
  (interactive)
  (let ((selected-files (dired-get-marked-files nil)))
    (if (and (= 1 (length selected-files)) (file-regular-p (car selected-files)))
      (let* ((full-path-midi-file (car selected-files)))
         (illustrate-midi-file full-path-midi-file))
      (progn
        (beep)
        (message "Cannot illustrate. You should only select a single midi file.")))))


; (defun process-laml-file-and-play ()
;   (interactive)
;   (let* ((laml-file-name (current-file))
;          (proper-laml-file-name-1 (file-name-proper laml-file-name))   ; with possible double name (icon name: x.icon)
;          (proper-laml-file-name-2 (file-name-proper (file-name-proper laml-file-name)))   ; without possible double name
;          (midi-file-name-1 (concat proper-laml-file-name-1 ".mid"))
;          (midi-file-name-2 (concat proper-laml-file-name-2 ".mid"))
;          (midi-file-path-1 (concat (current-directory) "midi-temp/" midi-file-name-1))
;          (midi-file-path-2 (concat (current-directory) "midi-temp/" midi-file-name-2)))        
;     (copy-file path-to-midi-error-file midi-file-path-1 t) 
;     (copy-file path-to-midi-error-file midi-file-path-2 t) 
;     (laml-sync-process-current-buffer)  
;     (message "A MIDI file has been generated in midi-temp. Now playing...")
;     (cond ((file-exists-p midi-file-path-1)   ; with double name
;            (play-midi-file midi-file-path-1))
;           ((file-exists-p midi-file-path-2)   ; without double name
;            (play-midi-file midi-file-path-2))
;           (t (error "Process-laml-file-and-play: Should not happen:")))))

(defun process-laml-file-and-play ()
  (interactive)
  (let* ((laml-file-name (current-file))
         (proper-laml-file-name (file-name-proper laml-file-name))   ; with possible double name (icon name: x.icon)
         (midi-file-name (concat proper-laml-file-name ".mid"))
         (midi-file-path (concat (current-directory) "midi-temp/" midi-file-name))
        )
    (ensure-directory-existence (current-directory) "midi-temp/")
    (copy-file path-to-midi-error-file midi-file-path t) 
    (laml-sync-process-current-buffer)    ; playing with (laml-sync-process-current-buffer "laml-source")
    (message "A MIDI file has been generated in midi-temp. Now playing...")
    (cond ((file-exists-p midi-file-path)   ; with possible double name
            (play-midi-file midi-file-path))
          (t (error "Process-laml-file-and-play: Should not happen:")))))

(defun process-laml-file-and-illustrate ()
  (interactive)
  (let* ((laml-file-name (current-file))
         (proper-laml-file-name (file-name-proper laml-file-name))   ; with possible double name (icon name: x.icon)
         (midi-file-name (concat proper-laml-file-name ".mid"))
         (midi-file-path (concat (current-directory) "midi-temp/" midi-file-name))
        )
    (ensure-directory-existence (current-directory) "midi-temp/")
    (copy-file path-to-midi-error-file midi-file-path t) 
    (laml-sync-process-current-buffer)  
    (message "A MIDI file has been generated in midi-temp. Now illustrating...")
    (cond ((file-exists-p midi-file-path)   ; with possible double name
            (illustrate-midi-file midi-file-path))
          (t (error "Process-laml-file-and-play: Should not happen:")))))


(defun process-laml-file-and-save (target-file-name)
  "Process midi laml file of the current buffer and save it in the file target-file-name.
target-file-name is assumed to be the full path to the target file name."
  (interactive "sName of target midi file (without extension): ")
  (let* ((kind (kind-of-midl-file-in-current-buffer)))
    (cond ((eq 'midi kind)      (process-laml-midi-file-and-save target-file-name))
          ((eq 'multi-pad kind) (process-laml-multi-pad-file-and-save target-file-name))
          ((eq 'style kind) (process-laml-style-file-and-save target-file-name))
          (t (error (concat "process-laml-file-and-save: Unknown kind of midl file" (symbol-to-string kind)))))))

(defvar default-icon-name "S910")

(defun process-laml-midi-file-and-save (target-file-name)
  (let* ((laml-file-name (current-file))
         (proper-laml-file-name (file-name-proper laml-file-name))   
         (temp-source-file-path (concat (current-directory) "midi-temp/" proper-laml-file-name "." "mid"))
         (target-file-path (concat (current-directory) target-file-name "." "mid")))
    (ensure-directory-existence (current-directory) "midi-temp/")
    (laml-sync-process-current-buffer)
    (copy-file temp-source-file-path target-file-path t)
    (message (concat "The file "  target-file-path  " has been created."))
))

(defun process-laml-multi-pad-file-and-save (target-file-name)
  (let* ((laml-file-name (current-file))
         (proper-laml-file-name (file-name-proper laml-file-name))
         (icon-name (file-name-extension proper-laml-file-name))  ; because laml-file-name has double extension - icon in the middle.
         (temp-source-file-path (concat (current-directory) "midi-temp/" proper-laml-file-name "." "mid"))
         (target-file-path 
             (if icon-name
                 (concat (current-directory) target-file-name "." icon-name "." "pad")
                 (concat (current-directory) target-file-name "." default-icon-name "." "pad"))))
    (ensure-directory-existence (current-directory) "midi-temp/")
    (laml-sync-process-current-buffer)
    (copy-file temp-source-file-path target-file-path t)
    (message (concat "The file "  target-file-path  " has been created."))))

(defun process-laml-style-file-and-save (target-file-name)
  (let* ((laml-file-name (current-file))
         (cur-dir (current-directory))
         (proper-laml-file-name (file-name-proper laml-file-name))
         (icon-name (file-name-extension proper-laml-file-name))  ; because laml-file-name has double extension - icon in the middle.
         (temp-source-file-path (concat cur-dir "midi-temp/" proper-laml-file-name "." "mid"))
         (target-file-path ; not used
             (if icon-name
                 (concat cur-dir target-file-name "." icon-name "." "sty")
                 (concat cur-dir target-file-name "." default-icon-name "." "sty")))
         (extension-path (concat cur-dir "midi-temp/" "extensions"))
        )
    (ensure-directory-existence cur-dir "midi-temp/")
    (laml-sync-process-current-buffer)  ; also makes the extensions file in midi-temp
    (let* ((explicit-extensions (file-read extension-path)) 
           (explicit-target-file-path (concat cur-dir target-file-name "." explicit-extensions)))
      (copy-file temp-source-file-path explicit-target-file-path t)
      (delete-file extension-path)
      (message (concat "The file "  explicit-target-file-path  " has been created."))) ))



(defun process-midl-file-source-target (source-file-path target-file-path)
  "Process the midi laml file in source-file-path and write the resulting midi file to target-file-path.
Both are full file paths. The processing is done via the laml temp directory, in order not to affect the directory 
that holds the source file."
  (let* ((file-name-ext (file-name-nondirectory source-file-path))
         (proper-file-name (file-name-proper file-name-ext))
         (source-file-path-in-temp-dir (concat laml-temp-file-path file-name-ext)))
    (copy-file source-file-path source-file-path-in-temp-dir t)
    (let* ((generated-midi-file-path (concat laml-temp-file-path "midi-temp/" proper-file-name "." "mid")) )
      (ensure-directory-existence laml-temp-file-path "midi-temp/")
      (copy-file path-to-midi-error-file generated-midi-file-path  t) 
      (laml-sync-process-file source-file-path-in-temp-dir 'silent)
      (delete-file source-file-path-in-temp-dir)
      (copy-file generated-midi-file-path target-file-path t)
      'done)))

; An ad hoc midl example processing function.
; The following functions, process-all-midl-files, generalizes it in a useful way.
(defun process-midl-example-files ()
  "Process all midl files in the current directory (such as a dired directory)
and write the resulting midi files to a specialized directory in the parent of the current directory.
Just right for the organization of the example files."
  (interactive)
  (let* ((dir (current-directory))
         (dir-path-list (file-path-to-path-list dir))
         (dir-name (car (last dir-path-list))) 
         (midl-files (filter (function (lambda (f) (equal (file-name-extension f) "midl"))) (directory-files dir t)))  ; full paths
         (target-dir (concat dir "../" dir-name "-midi" "/")))
    (ensure-directory-existence (concat dir "../") (concat dir-name "-midi"))
    (mapcar2
     (function
      (lambda (sfp n)
        (message "Processing file number %d: %s" n (file-name-nondirectory sfp))
        (process-midl-file-source-target sfp (concat target-dir (file-name-proper (file-name-nondirectory sfp)) "." "mid"))))
     midl-files (number-interval 1 (length midl-files)))
    (message "Processed %d midl files" (length midl-files))))

(defun process-all-midl-files (target-dir)
  "Process all midl files in the current directory (such as a dired directory)
and write the derived midi files in target-dir." 
  (interactive "DTarget directory: ")
  (let* ((dir (current-directory))
         (dir-path-list (file-path-to-path-list dir))
         (dir-name (car (last dir-path-list))) 
         (midl-files (filter (function (lambda (f) (equal (file-name-extension f) "midl"))) (directory-files dir t)))  ; full paths
        )
    (mapcar2
     (function
      (lambda (sfp n)
        (message "Processing file number %d: %s" n (file-name-nondirectory sfp))
        (process-midl-file-source-target sfp (concat target-dir (file-name-proper (file-name-nondirectory sfp)) "." "mid"))))
     midl-files (number-interval 1 (length midl-files)))
    (message "Processed %d midl files" (length midl-files))))


; ---------------------------------------------------------------------------------------------------
; Midl source stack
; Invariant: The Midl stack (of size n) is formed by the current buffer at the top, and
; a number of files 0.midl, ... n-1.midl in the "project directory".
; midl-source-file-stack is a stack (list) of the file-paths 0.midl ... n-1.midl.

(defvar midl-source-file-stack '())
(make-variable-buffer-local 'midl-source-file-stack)

(defvar midl-source-file-ghost-stack '())
(make-variable-buffer-local 'midl-source-file-ghost-stack)

(defun pop-and-reestablish-midl-file ()
  (interactive)
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (if (not (null midl-source-file-stack))
      (let* ((file-path (car midl-source-file-stack))
             (ghost-file-path (make-ghost-file-path file-path))
            )
        ; Write buffer to ghost file:
        (register-current-point (point))
        (write-region (point-min) (point-max) ghost-file-path)

        ; Update buffer from top of stack
        (erase-buffer)
        (insert-file-contents file-path)
        (buffer-reset-undo)
        (let ((registered-point (extract-registered-point)))
           (if registered-point
               (goto-char registered-point)))
        (save-buffer)

        ; Delete file corresponding to top of stack:
        (delete-file file-path)
        
        ; Update stack lists:
        (setq midl-source-file-ghost-stack (cons ghost-file-path midl-source-file-ghost-stack))
        (setq midl-source-file-stack (cdr midl-source-file-stack))

        ; Info:
;        (if (= (length midl-source-file-stack) 0)
;               (message (concat "The stack has been popped, and it has now reached its minimal height. "
;                                "Ghost files: %d.")  (length midl-source-file-ghost-stack))
;               (message "The stack has been popped. Remaining entries on the stack (including this buffer): %d. Ghost files: %d."
;                        (+ (length midl-source-file-stack) 1) (length midl-source-file-ghost-stack)))
        (message (concat "After pop: %d element(s) + %d ghost(s). " (comment-of-midl-file))
                        (+ (length midl-source-file-stack) 1) (length midl-source-file-ghost-stack))
      )
      (error (concat "The midl source file stack only contains the basis buffer, which cannot be popped. Nothing done"))))

(defun make-ghost-file-path (file-path)
  "Return the ghost-file-path of file-path.
A ghost file is a specially named file that is only saved for pop-undo purposes."
  (let ((proper-name-part (file-name-proper (file-name-nondirectory file-path)))
        (directory-file-path (file-name-directory file-path)))
    (concat directory-file-path proper-name-part "-" "ghost" "." "midl")))

(defun undo-pop-midl-stack ()
  (interactive)
  (if (>= (length midl-source-file-ghost-stack) 1)
      (let* ((ghost-file-path (car midl-source-file-ghost-stack))
             (file-path (make-non-ghost-file-path ghost-file-path)))

        ; Establish a new file from buffer - numbering comes derived top ghost file name:
        (write-region (point-min) (point-max) file-path)

        ; Establish buffer from ghost file:
        (erase-buffer)
        (insert-file-contents ghost-file-path)
        (buffer-reset-undo)
        (let ((registered-point (extract-registered-point)))
           (if registered-point
               (goto-char registered-point)))
        (save-buffer)

        ; Delete ghost file:
        (delete-file ghost-file-path)

        ; Update stack list:
        (setq midl-source-file-ghost-stack (cdr midl-source-file-ghost-stack))
        (setq midl-source-file-stack (cons file-path midl-source-file-stack))

        ; Info:
        (message (concat "After undo pop: %d element(s) + %d ghost(s). " (comment-of-midl-file))
                        (+ (length midl-source-file-stack) 1) (length midl-source-file-ghost-stack))
      )
      (message "Nothing to undo. Nothing done.")))

; (defun replicate-and-undo-all-pops ()
;   "Replicate current buffer of the absolute top of the stack, turning all ghost elements
; into real elements (undoing all pops)."
;   (interactive)
;   (while (not (null midl-source-file-ghost-stack))
;      (let* ((ghost-path (car midl-source-file-ghost-stack))
;             (non-ghost-path (make-non-ghost-file-path ghost-path)))
;        (rename-file ghost-path non-ghost-path t)      
;        (setq midl-source-file-ghost-stack (cdr midl-source-file-ghost-stack))
;        (setq midl-source-file-stack (cons non-ghost-path midl-source-file-stack))
;      )       
;   ))

(defun replicate-current-buffer-and-undo-all-pops ()
  "Replicate current buffer to the absolute top of the stack, turning all ghost elements
into real elements (undoing all pops)."
  (interactive)
  (let ((content-to-replicate (buffer-substring-no-properties (point-min) (point-max)))
        (current-project-name (file-name-proper (file-name-proper (current-file)))))

    ; Do as many undo-pops as we have elements on the ghost stack:
    (while (not (null midl-source-file-ghost-stack))
       (undo-pop-midl-stack))
    
    ; Push content-to-replicate on stack:
    (push-current-midl-buffer current-project-name (current-buffer))
    (erase-buffer)
    (insert content-to-replicate)
    (buffer-reset-undo)
    (save-buffer)
 
    ; Info:
    (message "Current buffer has been replicated to top of stack. Current height: %d. Number of ghosts: %d."
             ( + (length midl-source-file-stack) 1) (length midl-source-file-ghost-stack))
  )
)


(defun make-non-ghost-file-path (ghost-file-path)
  "Return the non-ghost file-path of ghost-file-path.
A ghost file is a specially named file that is only saved for pop-undo purposes."
  (let* ((proper-name-part (file-name-proper (file-name-nondirectory ghost-file-path)))
         (i (find-in-string proper-name-part ?-))       
         (directory-file-path (file-name-directory ghost-file-path)))
    (concat directory-file-path (substring proper-name-part 0 i) "." "midl")))


(defun push-current-midl-buffer (project-name buf)
 (save-excursion
  (set-buffer buf)
  (ensure-directory-existence (current-directory) (concat project-name "/") )
  (let ((temp-midl-file (concat (current-directory) (concat project-name "/") (int-to-string (length midl-source-file-stack)) "." "midl")))
    (register-current-point (point))
    (write-region (point-min) (point-max) temp-midl-file nil 0)
    (setq midl-source-file-stack (cons temp-midl-file midl-source-file-stack)))))

(defun midl-source-file-stack-info ()
  "Provide information about the midl source file stack"
  (interactive)
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (cond ((null midl-source-file-stack)
           (message "The midl stack only contains the basis source file. Stack height: 1. Ghost elements: %d" 
                    (length midl-source-file-ghost-stack)))
        (t (message "Height of midl source file stack: %d: This buffer and %d elements below it. Ghost elements: %d"
                     (+ 1 (length midl-source-file-stack))
                     (length midl-source-file-stack)
                     (length midl-source-file-ghost-stack)))))

(defun add-comment-to-midl-file (comment)
  "Add a comment to a midl file. The comment is a one-line comment stored as a Scheme comment at the first line of the file."
  (interactive (list (read-from-minibuffer "Midl file comment: " (comment-of-midl-file))))
  (set-comment-of-midl-file comment)
)

(defun display-comment-of-midl-file ()
  (interactive)
  (let ((comment (comment-of-midl-file)))
    (if (empty-string-p comment)
        (message "No comment of this midl file. Add one with M-x add-comment-to-midl-file or  C-c C-a")
        (message comment))))
        

(defun is-midl-stack-inconsistent()
  "Is the midl stack of the current project (implicitly determined from proper file name) inconsistent in relation to the
stack files in the project directory."
  (let* ((laml-file-name (current-file))
         (proper-laml-file-name (file-name-proper laml-file-name))
         (project-name (file-name-proper proper-laml-file-name))  ; removing possible double extension 
         (project-dir (concat (current-directory) project-name "/")))
    (and (null midl-source-file-stack)
         (or (file-exists-p (concat project-dir "/" "0" "." "midl")) 
             (file-exists-p (concat project-dir "/" "0-ghost" "." "midl")))))) 

(defun recover-midl-source-file-stack ()
  "Take the information in the project directory, and establish midl stack from it. 
Useful if you return to source file for which in another sessions have left information on the stack."
;  (interactive)
  (let ((current-project-name (file-name-proper (file-name-proper (current-file)))))
    (recover-middle-stack-from-project-directory current-project-name)))

(defun recover-middle-stack-from-project-directory (project-name)
  (let* ((project-dir (concat (current-directory) project-name "/")))
    (if (file-exists-p project-dir)
        (if (null midl-source-file-stack)
            (let* ((stack-files (filter 
                                  (function (lambda (x) 
                                            (let ((y (file-name-proper (file-name-nondirectory x))))
                                              (numeric-string y))))
                                  (directory-files (concat (current-directory) project-name "/"))))
                   (sorted-stack-files (sort stack-files (function (lambda (s1 s2) (or (equal s1 s2) (string-lessp s1 s2))))))

                   (ghost-files (filter 
                                  (function (lambda (x) 
                                            (let ((y (file-name-proper (file-name-nondirectory x))))
                                              (ghost-name-p y))))
                                  (directory-files (concat (current-directory) project-name "/"))))
                   (sorted-ghost-files (sort ghost-files (function (lambda (s1 s2) (or (equal s1 s2) (string-lessp s1 s2))))))
                  )
              (setq midl-source-file-stack 
                    (mapcar (function (lambda (f) (concat (current-directory) project-name "/" f))) (reverse sorted-stack-files)))
              (setq midl-source-file-ghost-stack 
                    (mapcar (function (lambda (f) (concat (current-directory) project-name "/" f))) sorted-ghost-files))
              (message "The midl stack has been reestablished. Current height (including current buffer): %s."
                       (+ 1 (length midl-source-file-stack))))
          (message "The stack already exists. Height (including current buffer) %s. Nothing done." (+ 1 (length midl-source-file-stack))))
      (message "Nothing to recover. Nothing done."))))

(defun reset-midl-source-file-stack ()
  "Get rid of earlier pushed midl files on the stack. That is, clear the stack entirely.
Deletes stack files in the project directory."
  (interactive)
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (if (ask-user "Do you really want to loose all stack frame, including the ghost elements?")
      (let ((current-project-name (file-name-proper (file-name-proper (current-file)))))
        (get-rid-of-ghost-stack)    
        (reset-midl-stack current-project-name))
      (message "Nothing Done.")))

(defun reset-midl-ghost-stack ()
  "Get rid all ghost elements."
  (interactive)
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (get-rid-of-ghost-stack)
  (message "Midl stack height (including this buffer): %d. Ghost files: %d."
           (+ (length midl-source-file-stack) 1) (length midl-source-file-ghost-stack))
)
    

(defun reset-midl-stack (project-name)
  (let* ((project-dir (concat (current-directory) project-name "/")))
    (setq midl-source-file-stack nil)
    (if (file-exists-p project-dir)
        (let ((stack-files (filter 
                                (function (lambda (x) 
                                            (let ((y (file-name-proper (file-name-nondirectory x))))
                                              (numeric-string y))))
                                (directory-files (concat (current-directory) project-name "/")))))
          (cond ((null stack-files) (message "No stack files in existing project directory. Nothing done."))
                ((not (null stack-files))
                   (mapcar (function 
                            (lambda (proper-file-name)
                              (delete-file (concat (current-directory) project-name "/" proper-file-name))))
                           stack-files)
                   (message "%s stack file(s) have been deleted. The midl stack now only consists of the current buffer."
                           (length stack-files)))))
        (message "Cannot find any project directory. Nothing done."))))


(defun push-copy-of-existing-midl-stack-entry (i)
  (interactive
    (list (if (= (length midl-source-file-stack) 0)  ; only possibility is current buffer = 0
              0
              (string-to-number
               (read-from-minibuffer 
                (format "Which entry do you wish to replicate: (0 = bottom, ..., %d = top = current buffer): " 
                        (length midl-source-file-stack)))))))
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (let* ((laml-file-name (current-file))
         (proper-laml-file-name (file-name-proper laml-file-name))
         (project-name (file-name-proper proper-laml-file-name))) ; eliminate possible double extension
    (cond ((= i (length midl-source-file-stack)) 
                   (push-current-midl-buffer project-name (current-buffer))
                   (message (concat "The existing buffer has been replicated on the midl stack.")))
          ((and (< i (length midl-source-file-stack)) (>= i 0))
              (push-current-midl-buffer project-name (current-buffer))
              (erase-buffer)
              (insert-file-contents (concat (current-directory) project-name "/" (int-to-string i) ".midl"))
              (buffer-reset-undo)
              (message "Element number %d from the stack has been copied to the top." i))
          (t (message "Illegal copying. Provide a number between 0 and %d. Nothing done." (length midl-source-file-stack))))))

(defun push-latest-temporary-selection-on-midl-stack ()
  (interactive)
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))

  (if (file-exists-p (concat (current-directory) "midi-temp"))
      (if (file-exists-p (concat (current-directory) "midi-temp/" "sel-temp.midl"))
          (let ((current-project-name (file-name-proper (file-name-proper (current-file)))))
            (get-rid-of-ghost-stack)
            (push-current-midl-buffer current-project-name (current-buffer))
            (erase-buffer)
            (insert-file-contents (concat (current-directory) "midi-temp/" "sel-temp.midl"))
            (buffer-reset-undo)
            (save-buffer)

            ; Info:
            (message "The latest temporary selection has been pushed on midl stack. Current height: %d. Number of ghosts: %d."
                     ( + (length midl-source-file-stack) 1) (length midl-source-file-ghost-stack))
          )
          (message "The temporary file sel-temp.midl does not exists in the midi-temp directory. Nothing done."))
      (message "The midi-temp directory does not exists. Nothing done."))
)

(defun push-external-midl-file-on-midl-stack (midl-file-path)
  (interactive "fPush which existing midl file: ")
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))

  (if (file-exists-p midl-file-path)
      (let ((current-project-name (file-name-proper (file-name-proper (current-file)))))
        (get-rid-of-ghost-stack)
        (push-current-midl-buffer current-project-name (current-buffer))
        (erase-buffer)
        (insert-file-contents midl-file-path)
        (buffer-reset-undo)
        (save-buffer)

                                        ; Info:
        (message "An external file has been pushed on midl stack. Current height: %d. Number of ghosts: %d."
                 ( + (length midl-source-file-stack) 1) (length midl-source-file-ghost-stack))
      )
    (message "The temporary file sel-temp.midl does not exists in the midi-temp directory. Nothing done."))
)




(defvar ciffer-list (list 48 49 50 51 52 53 54 55 56 57) "The list of chars that are ciffers: 0-9")

(defun numeric-string (str)
  (let* ((str-lst (string-to-list str))
         (non-numeric-elements (filter (function (lambda (ch) (not (member ch ciffer-list)))) str-lst)))
    (and (> (length str) 0) (null non-numeric-elements))))

(defun ghost-name-p (str)
  (and (>= (length str) 6)
       (equal "-ghost" (substring str (- (length str) 6) (length str))))) 

(defun get-rid-of-ghost-stack ()
  (while (not (null midl-source-file-ghost-stack))
    (delete-file (car midl-source-file-ghost-stack))
    (setq midl-source-file-ghost-stack (cdr midl-source-file-ghost-stack))))

(defun process-laml-file-save-and-push ()
  (interactive)
  (ensure-directory-existence (current-directory) "midi-temp/")
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (get-rid-of-ghost-stack)
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (let* ((kind (kind-of-midl-file-in-current-buffer)))
    (cond ((eq 'midi kind)      (process-laml-midi-file-save-and-push))
          ((eq 'multi-pad kind) (process-laml-multi-pad-file-save-and-push))
          ((eq 'style kind) (process-laml-style-file-save-and-push))
          (t (error (concat "process-laml-file-save-and-push: Unknown kind of midl file" (symbol-to-string kind)))))))

(defun process-laml-file-save-and-push-with-counter-transposition (ct)
  (interactive "nCounter transposition: ")
  (ensure-directory-existence (current-directory) "midi-temp/")
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (get-rid-of-ghost-stack)
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (let* ((kind (kind-of-midl-file-in-current-buffer)))
    (cond ((eq 'midi kind)      (process-laml-midi-file-save-and-push (number-to-string ct) ))
          ((eq 'multi-pad kind) (process-laml-multi-pad-file-save-and-push (number-to-string ct)))
          ((eq 'style kind) (laml-error "Cannot push with counter transposition on styles."))
          (t (error (concat "process-laml-file-save-and-push-with-counter-transposition: Unknown kind of midl file" 
                            (symbol-to-string kind)))))))

(defun process-laml-file-save-and-push-with-new-processing-mode (mode)
  (interactive 
     (list (completing-read "Processing mode for new top: " 
                   '(("absTime" . nil) ("deltaTime" . nil)) nil t)))
  (ensure-directory-existence (current-directory) "midi-temp/")
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (get-rid-of-ghost-stack)
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (let* ((kind (kind-of-midl-file-in-current-buffer)))
    (cond ((eq 'midi kind)      (process-laml-midi-file-save-and-push nil mode ))
          ((eq 'multi-pad kind) (process-laml-multi-pad-file-save-and-push nil mode))
          ((eq 'style kind) (process-laml-style-file-save-and-push mode))
          (t (error (concat "process-laml-file-save-and-push-with-new-processing-mode: Unknown kind of midl file" 
                            (symbol-to-string kind)))))))

; If provided, explicit-counter-transposition is a string
; If provided, new-processing-mode is either the the string "absTime" or "deltaTime"
(defun process-laml-midi-file-save-and-push (&optional explicit-counter-transposition new-processing-mode)
  (let* ((cur-buf (current-buffer))
         (laml-file-name (current-file))
         (proper-laml-file-name (file-name-proper laml-file-name))
         (project-name proper-laml-file-name) 
         (temp-midi-source-path (concat (current-directory) "midi-temp/" proper-laml-file-name "." "mid"))
         (temp-laml-target-path (concat (current-directory) "midi-temp/" "temp" "." "laml"))
         (processing-mode (if new-processing-mode new-processing-mode (find-processing-mode-of-current-buffer)))
         (counter-transposition (if explicit-counter-transposition explicit-counter-transposition (find-counter-transposition-of-current-buffer)))
        )

    (if (file-exists-p temp-laml-target-path) (delete-file temp-laml-target-path))  ; should not exist under normal circumstances
    (copy-file (concat laml-dir "styles/xml-in-laml/midi/" "templates/error-mid.tpl") temp-laml-target-path t)  ; rewritten if everything is OK.

    ; to avoid getting a wrong midi file later on:
    (if (file-exists-p temp-midi-source-path) (delete-file temp-midi-source-path))

    (message "Generating MIDI file from LAML file...")

    ; Process with eliminated target-format
    (let ((buf (current-buffer))
          (target-format (target-format-of-current-buffer)))

      ; Temporarily eliminate target-format form buffer:
      (if target-format (eliminate-target-format-of-current-buffer))
      (laml-sync-process-current-buffer) ; makes a midi file, in path temp-midi-source-path, in the midi-temp directory.
      (set-buffer buf)
      (if target-format (insert-target-format-in-current-buffer target-format))

      (push-current-midl-buffer project-name cur-buf)
      (message "Reprocessing  MIDI file in order to produce a fresh LAML (midl) file...")
      (sync-midi-file-to-laml-file processing-mode counter-transposition nil temp-midi-source-path temp-laml-target-path)
      (set-buffer cur-buf)
      (erase-buffer)
      (insert-file-contents temp-laml-target-path)
      (buffer-reset-undo)
      (if target-format (insert-target-format-in-current-buffer target-format))   ; insert original target-format in pushed buffer
      (delete-file temp-laml-target-path)
      (let ((buf (get-buffer "temp-processing-program.laml")))
        (if buf (kill-buffer buf))))
    
    (message "An evaluation of the buffer has been pushed. New stack height (including this buffer): %s."   (+ 1 (length midl-source-file-stack)))
))

(defun split-and-replicate-midl-file-by-channel (ch)
  (interactive "nChannel number (1..16): ") 

  (ensure-directory-existence (current-directory) "midi-temp/")
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (get-rid-of-ghost-stack)
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (let* ((kind (kind-of-midl-file-in-current-buffer)))
    (cond ((eq 'midi kind)     
             (let ((format-0-file-p (current-buffer-is-standard-midifile-of-format 0))   ; bool
                   (mode (mode-of-standard-midi-file-current-buffer))) ; symbol
               (if (and format-0-file-p (eq mode 'absTime))
                      (split-and-replicate-midl-format0-abstime-file-by-channel ch))))
          ((eq 'multi-pad kind) (error "Cannot split a multipad file"))
          ((eq 'style kind) (error "Cannot split a style file"))
          (t (error (concat "split-and-replicate-midl-file-by-channel: Unknown kind of midl file" 
                            (symbol-to-string kind)))))))

(defun split-and-replicate-midl-format0-abstime-file-by-channel (ch)
  "Push two new entries on the stack. One disregard channel ch, and the other focussing on channel ch"
  (let* ((cur-buf (current-buffer))
         (laml-file-name (current-file))
         (proper-laml-file-name (file-name-proper laml-file-name))
         (project-name proper-laml-file-name)
         (temp-midl-file-name "_temp")   ; The name of temporary file
         (temp-midl-file-path  (concat (current-directory) temp-midl-file-name ".midl"))   ; The path to the temporary file in the current directory
         (temp-midi-source-path (concat (current-directory) "midi-temp/" temp-midl-file-name "." "mid"))
         (temp-laml-target-path (concat (current-directory) "midi-temp/" "temp" "." "laml"))
         (processing-mode "absTime")
         (counter-transp (find-counter-transposition-of-current-buffer))
         (current-buffer-contents (buffer-substring-no-properties (point-min) (point-max)))
        )

    (if (file-exists-p temp-laml-target-path) (delete-file temp-laml-target-path)) ; should not exist under normal circumstances
    (copy-file (concat laml-dir "styles/xml-in-laml/midi/" "templates/error-mid.tpl") temp-laml-target-path t) ; rewritten if everything is OK.

                                        ; to avoid getting a wrong midi file later on:
    (if (file-exists-p temp-midi-source-path) (delete-file temp-midi-source-path))

    (message "Generating MIDI files from LAML file...")
                                        ; Process with eliminated target-format

    ; Make and push split with all by channel ch:    
    (if (file-exists-p temp-midl-file-path) (delete-file temp-midl-file-path))
    (make-other-channels-file-from-buffer-string current-buffer-contents ch temp-midl-file-path)  ; makes a source with appropriate channel filtering in temp-midl-file-path
    (message "Pushing file with all channels but %s" ch)
    (laml-sync-process-file temp-midl-file-path 'silent)                   ; produces a midi file in temp-midi-source-path
    (push-current-midl-buffer project-name cur-buf)
    (sync-midi-file-to-laml-file processing-mode counter-transp nil 
                                 temp-midi-source-path temp-laml-target-path)   ; parse to laml, and place laml file in temp-laml-target-path
    (set-buffer cur-buf)
    (erase-buffer)
    (insert (format ";;;; SPLIT: All channels but channel %s" ch)) (insert CR)
    (insert-file-contents temp-laml-target-path)
    (buffer-reset-undo)
    (delete-file temp-laml-target-path)
    (delete-file temp-midl-file-path)

    ; Make and push split with channel ch:
    (make-channel-file-from-buffer-string current-buffer-contents ch temp-midl-file-path)    ; makes a source with appropriate channel filtering in temp-midl-file-path
    (message "Pushing file with channel %s only" ch)
    (laml-sync-process-file temp-midl-file-path 'silent)              ; produces a midi file in temp-midi-source-path
    (push-current-midl-buffer project-name cur-buf)
    (sync-midi-file-to-laml-file processing-mode counter-transp nil 
                                 temp-midi-source-path temp-laml-target-path)   ; parse to laml, and place laml file in temp-laml-target-path
    (set-buffer cur-buf)
    (erase-buffer)
    (insert (format ";;;; SPLIT: Only channel %s" ch)) (insert CR)
    (insert-file-contents temp-laml-target-path)
    (buffer-reset-undo)
    (delete-file temp-laml-target-path)
    (delete-file temp-midl-file-path)

    (message "The buffer has been split by channel %s. Two new entries have been pushed. New stack height (including this buffer): %s."  
             ch  (+ 1 (length midl-source-file-stack)))
    ))

(defun make-channel-file-from-buffer-string (buffer-string ch target-path)
  "Make a file in target-path with filtering to channel ch. Return t if successful, else nil"
  (let* ((temp-buffer-name " midl-split")
         (buf (get-buffer-create temp-buffer-name))
        )
    (set-buffer buf)
    (erase-buffer)
    (insert buffer-string)

    (goto-char (point-min))
    (let ((res1 (search-forward "(StandardMidiFile" nil t)))
      (if res1
          (let ((res2 (search-forward "(MidiTrack" nil t)))
            (if res2
                (progn
                  (search-backward "(MidiTrack" nil t)
                  (forward-sexp 1)
                  (insert ")")
                  (goto-char res2)
                  (insert " ")
;                  (insert (format "(filter-messages (lambda (x) (and (ast? x) (or (non-channel-message? x) (= %s (midi 'channel x)))))" ch))
                  (insert (format "(filter-messages (lambda (x) (and (ast? x) (or (Meta? x) (equal? %s (midi 'channel x)))))" ch))

                  (if (file-exists-p target-path) (delete-file target-path))
                  (write-file target-path)
                  (kill-buffer buf)
                  t)
              nil))
            nil))))

(defun make-other-channels-file-from-buffer-string (buffer-string ch target-path)
  "Make a file in target-path with filtering to all other channels than ch. Return t if successful, else nil"
  (let* ((temp-buffer-name " midl-split")
         (buf (get-buffer-create temp-buffer-name))
        )
    (set-buffer buf)
    (erase-buffer)
    (insert buffer-string)

    (goto-char (point-min))
    (let ((res1 (search-forward "(StandardMidiFile" nil t)))
      (if res1
          (let ((res2 (search-forward "(MidiTrack" nil t)))
            (if res2
                (progn
                  (search-backward "(MidiTrack" nil t)
                  (forward-sexp 1)
                  (insert ")")
                  (goto-char res2)
                  (insert " ")
                  (insert (format "(filter-messages (lambda (x) (and (ast? x) (or (non-channel-message? x) (not (= %s (midi 'channel x))))))" ch))

                  (if (file-exists-p target-path) (delete-file target-path))
                  (write-file target-path)
                  (kill-buffer buf)
                  t)
              nil))
            nil))))
  
(defun merge-two-top-stack-entries ()
  "Merge the two stack entries which both are assumed to be the result of a split"
  (interactive)
  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (if (not (null midl-source-file-stack))
      (let* ((file-path (car midl-source-file-stack)))
        (if (is-split-situation file-path (current-buffer))
            (let* ((cur-buf (current-buffer))
                   (laml-file-name (current-file))
                   (proper-laml-file-name (file-name-proper laml-file-name))
                   (project-name proper-laml-file-name) 
                   (temp-midi-source-path (concat (current-directory) "midi-temp/" proper-laml-file-name "." "mid"))
                   (temp-midi-source-path-1 (concat (current-directory) "midi-temp/" "_temp-1" "." "mid"))
                   (temp-midi-source-path-2 (concat (current-directory) "midi-temp/" "_temp-2" "." "mid"))
                   (processing-mode (find-processing-mode-of-current-buffer))
                   (counter-transposition (find-counter-transposition-of-current-buffer))
                   (temp-laml-target-path (concat (current-directory) "midi-temp/" "temp" "." "laml"))
                  )
              (get-rid-of-ghost-stack)

              ; Process top of stack - the slave (with events belonging to a single channel):
              (laml-sync-process-current-buffer) ; makes a midi file, in path temp-midi-source-path, in the midi-temp directory.
              (message "Make midi file from the top of the stack...")
              (rename-file temp-midi-source-path temp-midi-source-path-1 t)

              ; Process entry just under top - the master:
              (set-buffer cur-buf)
              (erase-buffer)
              (insert-file-contents file-path)
              (laml-sync-process-current-buffer)  ; makes a midi file, in path temp-midi-source-path, in the midi-temp directory.
              (message "Make midi file from the entry just below the top of the stack...")

              (merge-midi-files temp-midi-source-path-1 temp-midi-source-path temp-midi-source-path-2) ; merges two first midi files to a midi file in temp-midi-source-path-2
              (message "Merging two midi files...")

              (sync-midi-file-to-laml-file processing-mode counter-transposition nil temp-midi-source-path-2 temp-laml-target-path)
              (message "Making resulting Midi LAML file...")
 
              (set-buffer cur-buf)
              (erase-buffer)
              (insert-file-contents temp-laml-target-path)
              (buffer-reset-undo)
              (save-buffer)

              ; Delete file corresponding to top of stack:
              (delete-file file-path)
              (setq midl-source-file-stack (cdr midl-source-file-stack))

              ; Cleaning up:
              (delete-file temp-midi-source-path-1)
              (delete-file temp-midi-source-path)
              (delete-file temp-midi-source-path-2)

              (message (concat "After merge: %d element(s) + %d ghost(s). " (comment-of-midl-file))
                       (+ (length midl-source-file-stack) 1) (length midl-source-file-ghost-stack)))
          (message "The top and the entry just below the top seem NOT to be a result of a split"))
      )
      (message (concat "The midl source file stack only contains the basis buffer, which cannot be popped. Nothing done"))))

(defun is-split-situation (master-file-path slave-buffer) 
  t)

(defun merge-midi-files (slave-path master-path output-path)
   (make-a-file-from-laml-template
       "temp-processing-program.laml"
       laml-temp-file-path
       "merge-master-and-slave-to-single-midi-file"
       'midi-laml-mode
       (list
        (list "FILE-PATH-1" master-path)
        (list "FILE-PATH-2" slave-path)
        (list "FILE-PATH-3" output-path)
       )
       (concat laml-dir "styles/xml-in-laml/midi/" "templates/"))  
    (laml-sync-process-file (concat laml-temp-file-path "temp-processing-program.laml") 'silent))


(defun target-format-of-current-buffer ()
 (save-excursion
   (goto-char (point-min))
   (let* ((target-format-reg-exp 
             (regular-expression  '(concat "'target-format" (one-or-more (char-set tab newline space return)) (concat "\"" (char-set "0" "1" "2") "\"") )))
          (target-format-match (re-search-forward target-format-reg-exp nil t)))
     (if target-format-match
         (match-string-no-properties 0)
         nil))))

(defun eliminate-target-format-of-current-buffer()
  (save-excursion
   (goto-char (point-min))
   (let* ((target-format-reg-exp 
             (regular-expression  '(concat "'target-format" (one-or-more (char-set tab newline space return)) (concat "\"" (char-set "0" "1" "2") "\"") )))
          (target-format-match (re-search-forward target-format-reg-exp nil t)))
     (if target-format-match
         (replace-match "")
         nil))))

(defun insert-target-format-in-current-buffer (target-format-string)
  (save-excursion
   (goto-char (point-min))
   (let* ((format-reg-exp 
             (regular-expression  '(concat "'format" (one-or-more (char-set tab newline space return)) (concat "\"" (char-set "0" "1" "2") "\"") )))
          (format-match (re-search-forward format-reg-exp nil t)))
     (if format-match
         (insert (concat " " target-format-string " ")) 
         (error "Could not locate a format attribute of current midl buffer.")))))

; If provided, explicit-counter-transposition is a string
; If provided, new-processing-mode is either the the string "absTime" or "deltaTime"
(defun process-laml-multi-pad-file-save-and-push (&optional explicit-counter-transposition new-processing-mode)
  (let* ((cur-buf (current-buffer))
         (laml-file-name (current-file))
         (proper-laml-file-name (file-name-proper laml-file-name))
         (project-name (file-name-proper proper-laml-file-name))  ; eliminate possible double extension
         (temp-midi-source-path-0 (concat (current-directory) "midi-temp/" proper-laml-file-name "." "mid"))
         (temp-midi-source-path (concat (current-directory) "midi-temp/" proper-laml-file-name "." "pad"))
         (temp-laml-target-path (concat (current-directory) "midi-temp/" "temp" "." "laml"))
         (processing-mode (if new-processing-mode new-processing-mode (find-processing-mode-of-current-buffer)))
         (counter-transposition (if explicit-counter-transposition explicit-counter-transposition (find-counter-transposition-of-current-buffer)))
        )
    (if (file-exists-p temp-laml-target-path) (delete-file temp-laml-target-path))  ; should not exist under normal circumstances
    (copy-file (concat laml-dir "styles/xml-in-laml/midi/" "templates/error-pad.tpl") temp-laml-target-path)  ; rewritten if everything is OK.

    ; to avoid getting a wrong midi file later on:
    (if (file-exists-p temp-midi-source-path) (delete-file temp-midi-source-path))

    (message "Generating MIDI multi-pad file from LAML file...")
    (laml-sync-process-current-buffer) ; gives -.mid file
    (rename-file temp-midi-source-path-0 temp-midi-source-path t)   ; now pad name. Important for subsequent midi-to-laml processing.
    (push-current-midl-buffer project-name cur-buf)
    (message "Reprocessing  PAD file in order to produce a fresh LAML (midl) file...")
    (sync-midi-file-to-laml-file processing-mode counter-transposition nil temp-midi-source-path temp-laml-target-path)
    (set-buffer cur-buf)
    (erase-buffer)
    (insert-file-contents temp-laml-target-path)
    (buffer-reset-undo)
    (delete-file temp-laml-target-path)
    (let ((buf (get-buffer "temp-processing-program.laml")))
      (if buf (kill-buffer buf)))
    
    (message "An evaluation of the buffer has been pushed. New stack height (including this buffer): %s."   (+ 1 (length midl-source-file-stack)       )) ))

; If provided, explicit-counter-transposition is a string
; If provided, new-processing-mode is either the the string "absTime" or "deltaTime"
(defun process-laml-style-file-save-and-push (&optional new-processing-mode)
  (let* ((cur-buf (current-buffer))
         (laml-file-name (current-file))
         (proper-laml-file-name (file-name-proper laml-file-name))
         (project-name (file-name-proper proper-laml-file-name))  ; eliminate possible double extension
         (temp-midi-source-path-0 (concat (current-directory) "midi-temp/" proper-laml-file-name "." "mid"))
         (temp-midi-source-path (concat (current-directory) "midi-temp/" proper-laml-file-name "." "sty"))
         (temp-laml-target-path (concat (current-directory) "midi-temp/" "temp" "." "laml"))
         (processing-mode (if new-processing-mode new-processing-mode (find-processing-mode-of-current-buffer)))
        )
    (if (file-exists-p temp-laml-target-path) (delete-file temp-laml-target-path))  ; should not exist under normal circumstances
    (copy-file (concat laml-dir "styles/xml-in-laml/midi/" "templates/error-style.tpl") temp-laml-target-path)  ; rewritten if everything is OK.

    ; to avoid getting a wrong midi file later on:
    (if (file-exists-p temp-midi-source-path) (delete-file temp-midi-source-path))

    (message "Generating MIDI style file from LAML file...")
    (laml-sync-process-current-buffer) ; gives -.mid file
    (rename-file temp-midi-source-path-0 temp-midi-source-path t)   ; now sty name. Important for subsequent midi-to-laml processing.
    (push-current-midl-buffer project-name cur-buf)
    (message "Reprocessing STYLE file in order to produce a fresh LAML (midl) file...")
    (sync-midi-file-to-laml-file processing-mode "0" nil temp-midi-source-path temp-laml-target-path)
    (set-buffer cur-buf)
    (erase-buffer)
    (insert-file-contents temp-laml-target-path)
    (buffer-reset-undo)
    (delete-file temp-laml-target-path)
    (let ((buf (get-buffer "temp-processing-program.laml")))
      (if buf (kill-buffer buf)))
    
    (message "An evaluation of the buffer has been pushed. New stack height (including this buffer): %s."   (+ 1 (length midl-source-file-stack)       )) ))


(defun kind-of-midl-file-in-current-buffer ()
 "By looking at the midl midi laml source in the current buffer, determine its kind
Return a symbol such as midi or multi-pad. Return nil for unknown type."
 (save-excursion
   (goto-char (point-max))

   ; Compensate for registered point at the end of the buffer:
   (beginning-of-line 1)
   (if (looking-at ";;; ")
       (backward-char 1)
       (goto-char (point-max)))

   (backward-sexp 1)
   (cond ((looking-at "(StandardMidiFile") 'midi)
         ((looking-at "(TyrosMultiPad") 'multi-pad)
         ((looking-at "(TyrosStyle") 'style)
         (t nil))))
              

; --------------------------------------------------------------------------------------------------------------
; Base lining, project directory, and midi stack

; IMPORTANT INVARIANT: The baseline files must always be numbered from 1 to n.
; No holes can exist among baseline files. Do not manually delete a baseline file in the beginnig of middle of the range of
; baseline files.

(defun create-baseline-of-current-midl-source-file (explanation)
  "Create a new baseline (snapshot) of the current midl source file and save it permanently in
the implicitly determined project directory"
  (interactive "sExplain the role of this baseline: ")
  (let ((current-project-name (file-name-proper (file-name-proper (current-file)))))
    (ensure-directory-existence (current-directory) (concat current-project-name "/") )
    (let* ((existing-base-line-files (find-existing-baseline-files (current-directory) current-project-name))
           (next-baseline-number (+ (length existing-base-line-files) 1)))
      (let ((target-path (concat (current-directory) current-project-name "/" "baseline" "-" 
                                 (number-to-string next-baseline-number) "." "midl"))
            (target-explanation-path
                         (concat (current-directory) current-project-name "/" "baseline" "-" 
                                 (number-to-string next-baseline-number) "-" "explanation" )))
        (write-region (point-min) (point-max) target-path nil 0)
        (if (> (length explanation) 0)
            (write-file-with-initial-contents explanation target-explanation-path))
        (message "The file baseline-%d bas been created in the project directory. Never delete it again!"  next-baseline-number)))) )

(defun write-file-with-initial-contents (str path)
  "Create a file with str as initial contents. The path of the file is path."
  (let ((new-buffer (create-file-buffer path)))
    (set-buffer new-buffer)
    (erase-buffer)
    (insert str)
    (write-file path)
    (kill-buffer new-buffer)))

(defun find-existing-baseline-files (current-dir project-name)
  (filter 
   (function (lambda (x) 
               (let ((fnp (file-name-proper (file-name-nondirectory x)))
                     (fne (file-name-extension (file-name-nondirectory x))))
                 (and 
                  (not (or (equal fnp ".") (equal fnp "..")))
                  (> (length fnp) 8) 
                  (equal "baseline" (substring fnp 0 8))
                  (equal "midl" fne)))))
   (directory-files (concat current-dir project-name "/"))))

(defun push-baseline-on-midl-stack (baseline-number)
  "Push an existing baseline file onto the midl stack"
  (interactive
    (let* ((current-project-name (file-name-proper (file-name-proper (current-file))))
           (count (length (find-existing-baseline-files (current-directory) current-project-name))))
     (list (if (= count 1) ; only one possibility
               1
               (string-to-number
                (read-from-minibuffer 
                 (format "Which baseline entry between 1 and %d do you wish to push: " 
                         count )))))))

  (let* ((current-project-name (file-name-proper (file-name-proper (current-file))))
         (baseline-source-path 
          (concat (current-directory) current-project-name "/" "baseline" "-" (number-to-string baseline-number) ".midl"))
       )
    (if (file-exists-p baseline-source-path)
        (progn
          (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
          (push-current-midl-buffer current-project-name (current-buffer))
          (erase-buffer)
          (insert-file-contents baseline-source-path)
          (buffer-reset-undo)
          (message "Baseline file number %d has been pushed on the midl stack" baseline-number))
        (message "The file baseline-%d.midl does not exist in the project directory %s." baseline-number current-project-name))))

(defun baseline-present-info ()
  (interactive)
  (let ((info-buffer (if (get-buffer "*BaselineInfo*") (get-buffer "*BaselineInfo*") (generate-new-buffer "*BaselineInfo*")))
        (current-project-name (file-name-proper (file-name-proper (current-file))))
        (current-dir (current-directory))  ; bound before info buffer is created.
       )
    (delete-other-windows)
    (split-window-vertically)
    (show-buffer (other-window 1) info-buffer)
    (set-buffer info-buffer)
    (toggle-read-only -1)
    (erase-buffer)

    (insert "BASELINE FILE INFO.") (insert CR) (insert CR)
    (insert "Baseline files are snapshots of midl files, which have been saved permanently in the project directory.") (insert CR) 
    (insert (concat "The project directory is located in: " current-dir current-project-name "/" )) (insert CR) (insert CR) 

    (let ((base-line-files (find-existing-baseline-files current-dir current-project-name)))
      (cond ((and (file-exists-p (concat current-dir current-project-name "/")) (not (null base-line-files)))
               (mapcar 
                (function
                 (lambda (blf)
                   (insert blf) (insert ": ")
                   (let ((expl-path (concat current-dir current-project-name "/" (file-name-proper blf) "-" "explanation"))) 
                     (if (file-exists-p expl-path)
                         (progn (let ((res (insert-file-contents expl-path))) ; point is located before insertion. Therefore forward-char ...
                                  (forward-char (cadr res)))
                                (insert CR))
                       (progn (insert "No explanation available.") (insert CR))))))
                base-line-files ))
            ((and (file-exists-p (concat current-dir current-project-name "/")) (null base-line-files))
              (insert (concat "There are no baseline files in the project directory " current-project-name)))

            (t (insert (concat "There is no directory for the project " current-project-name))))) 

    (toggle-read-only 1)
    (set-buffer-modified-p nil)
    (other-window 1)))


; ---------------------------------------------------------------------------------------------------------------
; Score drawing

(defvar score-counter-transpostion 0)
(make-variable-buffer-local 'score-counter-transpostion)

; Decorate midi messages with sores between the points p and m.
; When called interactively, p and m is supposed to be point and mark.
(defun add-score-comments-of-region (p m)
 "Decorate the selected region with score comments. All channels are decorated."
 (interactive "r")
 (display-message-or-buffer "Adding scores to all channels of the selected region...")
 (save-excursion
   (let* ((p0 (min p m))
          (pn (max p m)))
     (setq score-counter-transpostion (string-to-int (find-counter-transposition-of-current-buffer)))
     (goto-char pn) (beginning-of-line 1)
     (re-search-backward note-on-reg-exp p0 t) (beginning-of-line 1)
     (while (and (> (point) 1) (>= (point) p0))
       (message (concat "Progress: " (progress pn p0 (point))))
       (decorate-midi-message-line-with-score t) (beginning-of-line 1)
       (let ((back-search (re-search-backward note-on-reg-exp p0 t)))
         (if back-search
             (beginning-of-line 1)
             (goto-char (- p0 1)))))))
 (message "Scores added to region."))

(defun add-score-comments-to-bars (start-end-bar-string)
  "Add score comments to a number of bars"
  (interactive "sFrom bar number to bar number (sep. by space), one bar number, or the current bar: ")
  (let* ((start-bar-number-0 (first-number-of-string start-end-bar-string))
         (start-bar-number (if start-bar-number-0 start-bar-number-0 (current-bar-number)))
         (end-bar-number (last-number-of-string start-end-bar-string))
         (start-p (locate-initial-bar start-bar-number))
         (end-p   (if end-bar-number (locate-final-bar end-bar-number) (locate-final-bar start-bar-number)))
        )
    (add-score-comments-of-region start-p end-p)))


(setq note-on-reg-exp (regular-expression '(concat "(" (zero-or-more " ") "NoteOn")))

; Decorate midi messages with sores between the points p and m.
; When called interactively, p and m are supposed to be point and mark.
; If channel is t, add scores for all channels. Else only for a specific channel.
(defun do-add-score-comments-of-region-channel (channel p m)
 "Decorate the selected region, and the given channel with score comments"
 (save-excursion
   (let* ((p0 (min p m))
          (pn (max p m)))
     (setq score-counter-transpostion (string-to-int (find-counter-transposition-of-current-buffer)))
     (goto-char pn) (beginning-of-line 1)
     (re-search-backward note-on-reg-exp p0 t) (beginning-of-line 1)
     (while (and (> (point) 1) (>= (point) p0))
       (message (concat "Progress: " (progress pn p0 (point))))
       (decorate-midi-message-line-with-score channel) (beginning-of-line 1)
       (let ((back-search (re-search-backward note-on-reg-exp p0 t)))
         (if back-search
             (beginning-of-line 1)
             (goto-char (- p0 1))))))))

(defun add-score-comments-of-form ()
  "Decorate the selected form with score comments. You are supposed to select a form by placing point at the initial parenthesis."
  (interactive)
  (message "Adding scores to all channels in form...")
  (save-excursion
    (let ((p0 (point)))
      (forward-sexp 1)
      (let ((pn (point)))
        (add-score-comments-of-region p0 pn))))
  (message "Scores added to form"))

(defun add-score-comments-of-form-channel (ch)
  "Decorate the selected form with score comments. You are supposed to select a form by placing point at the initial parenthesis."
  (interactive "nAdd score to which channel in form: ")
  (message "Adding scores to channel " (number-to-string ch) " in form...")
  (save-excursion
    (let ((p0 (point)))
      (forward-sexp 1)
      (let ((pn (point)))
        (do-add-score-comments-of-region-channel ch p0 pn))))
  (message "Scores added to form"))

(defun add-score-comments-of-region-channel (ch)
  "Decorate a given channel in the selected region with score comments."
  (interactive "nAdd score to which channel in form: ")
  (message "Adding scores to channel " (number-to-string ch) " in form...")
  (save-excursion
    (do-add-score-comments-of-region-channel ch (point) (mark)))
  (message "Scores added to form"))

(defun add-score-comments-of-bar-channel (ch start-end-bar-string)
  "Add score comments to a number of bars, only a single channel"
  (interactive "nAdd score to which channels: 
sFrom bar number to bar number (sep. by space), one bar number, or the current bar: ")
  (let* ((start-bar-number-0 (first-number-of-string start-end-bar-string))
         (start-bar-number (if start-bar-number-0 start-bar-number-0 (current-bar-number)))
         (end-bar-number (last-number-of-string start-end-bar-string))
         (start-p (locate-initial-bar start-bar-number))
         (end-p   (if end-bar-number (locate-final-bar end-bar-number) (locate-final-bar start-bar-number)))
        )
    (do-add-score-comments-of-region-channel ch start-p end-p)))


(defun remove-score-comments-of-region (p m)
 "Remove the score comments of the selected region"
 (interactive "r")
 (save-excursion
   (let* ((p0 (min p m))
          (pn (max p m)))
     (goto-char pn)
     (while (> (point) p0)
       (message (concat "Progress: " (progress pn p0 (point))))
       (let ((search-res (re-search-backward score-comment-line-reg-exp p0 t)))
          (if search-res 
              (kill-line-not-cr)
              (goto-char p0)))))))

(defun remove-score-comments-of-bars (start-end-bar-string)
  (interactive "sFrom bar number to bar number (sep. by space), one bar number, or the current bar: ")
  (let* ((start-bar-number-0 (first-number-of-string start-end-bar-string))
         (start-bar-number (if start-bar-number-0 start-bar-number-0 (current-bar-number)))
         (end-bar-number (last-number-of-string start-end-bar-string))
         (start-p (locate-initial-bar start-bar-number))
         (end-p   (if end-bar-number (locate-final-bar end-bar-number) (locate-final-bar start-bar-number)))
        )
    (remove-score-comments-of-region start-p end-p)))

(defun kill-line-not-cr ()
  (while (not (looking-at-chars (point) '(10 13))) (delete-char 1)))

(setq score-comment-line-reg-exp   (regular-expression '(or "; " "; o" "; <" )))

(defun remove-score-comments-of-form ()
  "Remove the score comments from form. You are supposed to select a form by placing point at the initial parenthesis."
  (interactive)
  (save-excursion
    (let ((p0 (point)))
      (forward-sexp 1)
      (let ((pn (point)))
        (remove-score-comments-of-region p0 pn))))
  (message "Removed"))

; OLD version
; (defun decorate-midi-message-line-with-score (channel)
;   (let ((channel-of-mes (channel-of-message-on-current-line)))
;      (if (or (eq channel t) (and channel-of-mes (= channel channel-of-mes)))
;          (let* ((kind-of-mes (kind-of-message-on-current-line)))
;            (if (and kind-of-mes (equal kind-of-mes "NoteOn"))
;                (let ((note-number (note-number-on-current-line)))
;                  (insert-comment-at-column-number score-start-column (draw-score-line note-number))))))))

; Decorate NoteOn message current line with score comment.
; Assume as a precondition that there is a NoteOn message on the current line.
; Assume also that point is at the beginning of the line.             
; Channel can be a number, or t for all channels.
(defun decorate-midi-message-line-with-score (channel)
  (let ((channel-of-mes (channel-of-message-on-current-line)))
     (if (or (eq channel t) (and channel-of-mes (= channel channel-of-mes)))
         (let ((note-number (note-number-on-current-line)))
           (insert-comment-at-column-number score-start-column (draw-score-line note-number))))))

(defvar lower-c-note-number 60)
(make-variable-buffer-local 'lower-c-note-number)
; (setq lower-c-note-number 60)    ; c4 = 60

(defun set-c-value (c-value)
 (interactive)
  (let ((before lower-c-note-number))
    (setq lower-c-note-number c-value)
    (message (concat "Lower C value changed from " (int-to-string before) " to " (int-to-string c-value)))))

(defun report-lower-c ()
  (interactive)
  (message (concat "Lower C in score comment has value " (int-to-string lower-c-note-number))))

(defvar score-start-column  135)
(make-variable-buffer-local 'score-start-column)
(setq score-start-column  135)

; Return a score line for note-number.
(defun draw-score-line (note-number)
  (let ((ct-note-number (+ score-counter-transpostion note-number)))   ; counter-transposed node number
    (concat
      (draw-dot       (+ lower-c-note-number -10) ct-note-number  'min) ; d, d#
      (draw-between-1 (+ lower-c-note-number -8) ct-note-number)   ; e    
      (draw-dot       (+ lower-c-note-number -7) ct-note-number)   ; f, f#
      (draw-between   (+ lower-c-note-number -5) ct-note-number)   ; g, g#    
      (draw-dot       (+ lower-c-note-number -3) ct-note-number)   ; a, a#
      (draw-between-1 (+ lower-c-note-number -1) ct-note-number)   ; h      

      (draw-line      (+ lower-c-note-number 0) ct-note-number)    ; c, c#
      (draw-between   (+ lower-c-note-number 2) ct-note-number)    ; d, d#
      (draw-line-1    (+ lower-c-note-number 4) ct-note-number)    ; e
      (draw-between   (+ lower-c-note-number 5) ct-note-number)    ; f, f#
      (draw-line      (+ lower-c-note-number 7) ct-note-number)    ; g, g#
      (draw-between   (+ lower-c-note-number 9) ct-note-number)    ; a, a#
      (draw-line-1    (+ lower-c-note-number 11) ct-note-number)   ; h
      (draw-between   (+ lower-c-note-number 12) ct-note-number)   ; c, c#
      (draw-line      (+ lower-c-note-number 14) ct-note-number)   ; d, d#
      (draw-between-1 (+ lower-c-note-number 16) ct-note-number)   ; e

      (draw-dot       (+ lower-c-note-number 17) ct-note-number)    ; f, f#
      (draw-between   (+ lower-c-note-number 19) ct-note-number)    ; g, g#
      (draw-dot       (+ lower-c-note-number 21) ct-note-number)    ; a, a#
      (draw-between-1 (+ lower-c-note-number 23) ct-note-number)   ; h
      (draw-dot       (+ lower-c-note-number 24) ct-note-number 'max) ; c, c#
    )))

(defun draw-line (low-value note-number &optional limit)
  (cond ((and (eq limit 'min) (< note-number low-value)) "<")
        ((= low-value note-number) "o")
        ((= (+ low-value 1) note-number) "=")
        ((and (eq limit 'max) (> note-number (+ low-value 1))) ">")
        (t "|")))

(defun draw-line-1 (low-value note-number &optional limit)
  (cond ((= low-value note-number) "o")
        ((and (eq limit 'max) (> note-number (+ low-value 1))) ">")
        (t "|")))

(defun draw-dot-1 (low-value note-number &optional limit)
  (cond ((= low-value note-number) "o")
        ((and (eq limit 'max) (> note-number low-value)) ">")
        (t ".")))

(defun draw-dot (low-value note-number &optional limit)
  (cond ((and (eq limit 'min) (< note-number low-value)) "<")
        ((= low-value note-number) "o")
        ((= (+ low-value 1) note-number) "=")
        ((and (eq limit 'max) (> note-number (+ low-value 1))) ">")
        (t ".")))

(defun draw-between (low-value note-number &optional limit)
  (cond ((= low-value note-number) "o")
        ((= (+ low-value 1) note-number) "=")
        ((and (eq limit 'max) (> note-number (+ low-value 1))) ">")
        (t " ")))

(defun draw-between-1 (low-value note-number &optional limit)
  (cond ((= low-value note-number) "o")
        ((and (eq limit 'max) (> note-number low-value)) ">")
        (t " ")))

; Return the column number of position p (such as (point))
(defun column-number-of (p)
  (save-excursion
    (beginning-of-line 1)
    (- p (point))))

; Insert comment-txt a column number cn at the current line
(defun insert-comment-at-column-number (cn txt)
  (delete-existing-comment-at-line)
  (end-of-line 1)
  (let ((eol-pos (column-number-of (point))))
    (cond ((<= eol-pos cn)
              (insert (make-string (- cn eol-pos) 32))
              (insert "; ")
              (insert txt))
          (t  (insert "")))))

(defun delete-existing-comment-at-line ()
 (save-excursion
  (end-of-line)
  (while (is-inside-comment)
     (backward-delete-char-untabify 1))
  (backward-char 1)
  (while (looking-at-chars (point) '(32))
     (delete-char 1) (backward-char 1))))


; point in front of message form.
; Return the kind of midi message at the current line. A string or nil.
(defun kind-of-message-on-current-line ()
  (save-excursion
    (let ((p0 (point))
          (eol-p (progn (end-of-line 1) (point))))
      (goto-char p0)
      (if (eobp)
          nil
          (progn
            (while (and (not (eobp)) (not (looking-at-chars (point) '(40)))) (forward-char 1)) ; goto start paren
            (if (or (eobp) (> (point) eol-p))
                nil
              (progn
                (forward-char 1)
                (if (looking-at-chars (point) '(40)) ; another start paren.
                    nil
                  (prog-name-under-point)))))))))

; point in front of message form.
; Return the channel of message at the current line. A number or nil
(defun channel-of-message-on-current-line ()
  (save-excursion
    (let ((p0 (point))
          (eol-p (progn (end-of-line 1) (point))))
      (goto-char p0)
      (if (eobp)
          nil
          (let ((res (search-forward "'channel" eol-p t)))
            (if res
                (progn
                  (while (looking-at-chars (point) '(32 34))   ; space, quote
                     (forward-char 1))
                  (let ((channel-string (prog-name-under-point)))
                    (string-to-number channel-string)))
                nil))))))


; point in front of NoteOn message form.
(defun note-number-on-current-line ()
 (save-excursion
   (search-forward "'note" nil t)
   (while (not (looking-at-chars (point) '(34))) (forward-char 1))  ; goto "
   (forward-char 1)
   (let ((str (prog-name-under-point)))
     (string-to-number str))))

; ---------------------------------------------------------------------------------------------------------------
; Progress feedback.

; actual value goes from from-value to to-value. Display percentage done progress.
(defun progress (from-value to-value actual-value)
  (concat (number-to-string (truncate (* (/ (float (abs (- actual-value from-value))) (abs (- from-value to-value))) 100))) "%% done..."))

; ---------------------------------------------------------------------------------------------------------------

(defun find-processing-mode-of-current-buffer()
  "Return a string. Defaults to absTime."
  (save-excursion
    (goto-char (point-min))
    (let ((search-res (search-forward "'mode" nil t)))
      (if search-res
          (progn
            (search-forward "\"" nil t)
            (prog-name-under-point))
          "absTime"))))

(defun guess-processing-mode (start-pos end-pos)
  "Guess the processing mode from the first midi event in the [start-pos end-pos] interval. 
start and end are buffer positions. Return a string, either absTime or deltaTime. Defaults to deltaTime."
  (save-excursion
    (let ((delta-search (progn (goto-char start-pos)  (search-forward "deltaTime" end-pos t)))
          (abs-search   (progn (goto-char start-pos) (search-forward "'absTime" end-pos t)))
         )
      (cond ((and delta-search abs-search)
               (if (< delta-search abs-search) "deltaTime" "absTime"))
            (delta-search "deltaTime")
            (abs-search "absTime")
            (t "deltaTime")))))

  

(defun find-pulses-per-quarter-note-of-current-buffer()
  "Return the number of pulses per quarter note of the current buffer. An integer.
Return nil if the number of pulses per quarter note cannot be determined"
  (cond ((or (current-buffer-is-standard-midifile) (current-buffer-is-style-format))
            (save-excursion
              (goto-char (point-min))
              (let ((search-res (search-forward "'pulsesPerQuarterNote" nil t)))
                (if search-res
                    (progn
                      (search-forward "\"" nil t)
                      (string-to-number (prog-name-under-point)))
                  nil))))
        ((current-buffer-is-multipad-format)
            1920)
        (t (error "Unknow format of midi laml file: Should be StandardMidiFile, TyrosMultiPad, or TyrosStyle"))))


; Return a string. Defaults to "0"
(defun find-counter-transposition-of-current-buffer()
  (save-excursion
    (goto-char (point-min))
    (let ((search-res (search-forward "'counterTransposition" nil t)))
      (if search-res
          (progn
            (search-forward "\"" nil t)
            (prog-name-under-point))
          "0"))))

; ---------------------------------------------------------------------------------------------------------------
; Selective Midi Playing. Format 0 stuff.

(defvar selection-begin nil "The start point of the previous selection")   ; not used right now
(defvar selection-end nil "The end point of the previous selection")       ; not used right now
(defvar last-temporary-midi-file-path nil "The full path to the previous midi file selection")
(defvar previous-midi-step-length nil "The length of the previous midi step. An integer.") 

(make-variable-buffer-local 'selection-begin)
(make-variable-buffer-local 'selection-end)
(make-variable-buffer-local 'last-temporary-midi-file-path)
(make-variable-buffer-local 'previous-midi-step-length)

(defun process-selected-track-and-play ()
  (interactive)
  (do-selected-track-and-play 'play)
)

(defun process-selected-track-and-illustrate ()
  (interactive)
  (do-selected-track-and-play 'illustrate)
)


(defun do-selected-track-and-play (command)
  "Point is assumed to be at the initial parenthesis of a MidiTrack form.
Wrap the selected track of a midl file into a proper context, process it, and do it.
Command is either the symbol play or illustrate.
Works for format 0 as well as format 1 files."

  (if (looking-at "(MidiTrack")
      (save-excursion
      (let ((start-p (first-noteon-in-track))
            (end-p (last-noteon-in-track)))
        (setq selection-begin start-p) (setq selection-end end-p) 

        (goto-char start-p)   ; to allow process-laml-file-and-present-from-to to locate the appropriate context from this position
        (process-laml-file-and-present-from-to t start-p end-p command)
      ))
      (progn
        (beep)
        (message "You must select the initial parenthesis of a MidiTrack in order to use this command."))))



; Assume point is at initial parenthesis in MidiTrack form
(defun first-noteon-in-track ()
  (save-excursion
    (search-forward "(NoteOn" nil t)
    (beginning-of-line 1)
    (point)))

; Assume point is at initial parenthesis in MidiTrack form
(defun last-noteon-in-track ()
  (save-excursion
    (forward-sexp 1)
    (search-backward "(NoteOn" nil t)
    (beginning-of-line 1) (forward-sexp 1)
    (point)))
  

(defun process-selected-laml-file-and-play ()
  "Wrap the selected part of a midl file into a proper context, process it, and play it.
Works for format 0 as well as format 1 files."
  (interactive)
  (let ((start-p (min (point) (mark)))
        (end-p (max (point) (mark))))
    (setq selection-begin start-p) (setq selection-end end-p) 
    (process-laml-file-and-present-from-to t start-p end-p 'play)))

(defun process-selected-bars-laml-file-and-play (start-end-bar-string)
  "Wrap the selected bars of a midl file into a proper context, process it, and play it.
Works for format 0 files as well as format 1 files."
  (interactive "sPlay from bar number to bar number (sep. by space), only one bar number, or the current bar: ")
  (let* ((start-bar-number-0 (first-number-of-string start-end-bar-string))
         (start-bar-number (if start-bar-number-0 start-bar-number-0 (current-bar-number)))
         (end-bar-number (last-number-of-string start-end-bar-string))
         (start-p (locate-initial-bar start-bar-number))
         (end-p   (if end-bar-number (locate-final-bar end-bar-number) (locate-final-bar start-bar-number)))
        )
;    (goto-char start-p)
    (setq selection-begin start-p) (setq selection-end end-p) 
    (process-laml-file-and-present-from-to t start-p end-p 'play)))

(defun process-current-bar-and-play (number-of-bars-0)
  "Wrap the current bar, and depending of the prefix argument a number of additional bars,
of a midl file into a proper context, process it, and play it.
Works for format 0 files as well as format 1 files."
  (interactive "p")
  (let* ((number-of-bars (if number-of-bars-0 number-of-bars-0 1))
         (start-bar-number (current-bar-number))
         (end-bar-number (+ (current-bar-number) (- number-of-bars 1)))
         (start-p (locate-initial-bar start-bar-number))
         (end-p   (if end-bar-number (locate-final-bar end-bar-number) (locate-final-bar start-bar-number)))
        )
    (if (= start-bar-number end-bar-number)
        (message "Playing %s in bar number %d..." (channel-info-string channels-to-play) start-bar-number)
        (message "Playing %s from bar %d to bar %d..." (channel-info-string channels-to-play) start-bar-number end-bar-number))
    (setq selection-begin start-p) (setq selection-end end-p) 
    (process-laml-file-and-present-from-to (channel-list-as-string channels-to-play) start-p end-p 'play)
    ))

(defun process-from-current-bar-to-end-and-play ()
  (interactive)
  (let* ((max-bar (find-max-bar-number))
         (remaining-bar-numbers (+ (- max-bar (current-bar-number)) 1))
        )
    (process-current-bar-and-play remaining-bar-numbers)  ; process and play from current bart number and to end 
  )
)

(defun goto-next-bar-and-process-current-bar-and-play ()
  "Goto next bar. Wrap that bar into a proper context, process it, and play number-of-bars-0 bars (defaults to a single bar).
Works for format 0 files as well as format 1 files."
  (interactive)
  (let ((start-bar-number (current-bar-number)))
      (goto-bar (+ start-bar-number 1))
      (if play-after-navigate
         (progn 
           (message "Playing %s in bar number %d..." (channel-info-string channels-to-play) (+ start-bar-number 1))
           (let* ((start-bar-number (current-bar-number))
                  (end-bar-number (current-bar-number))
                  (start-p (locate-initial-bar start-bar-number))
                  (end-p   (if end-bar-number (locate-final-bar end-bar-number) (locate-final-bar start-bar-number)))
                  )
             (setq selection-begin start-p) (setq selection-end end-p) 
             (process-laml-file-and-present-from-to (channel-list-as-string channels-to-play) start-p end-p 'play)))
         (message "Now at the beginning of bar %d." (+ start-bar-number 1)))))         

(defun goto-previous-bar-and-process-current-bar-and-play ()
  "Goto previous bar. Wrap that bar into a proper context, process it, and play it.
Works for format 0 files as well as format 1 files."
  (interactive)
  (let ((start-bar-number (current-bar-number)))
      (goto-bar (- start-bar-number 1))

      (if play-after-navigate
         (progn 
           (message "Playing %s in bar number %d..." (channel-info-string channels-to-play) (- start-bar-number 1))
           (let* ((start-bar-number (current-bar-number))
                  (end-bar-number (current-bar-number))
                  (start-p (locate-initial-bar start-bar-number))
                  (end-p   (if end-bar-number (locate-final-bar end-bar-number) (locate-final-bar start-bar-number)))
                  )
             (setq selection-begin start-p) (setq selection-end end-p) 
             (process-laml-file-and-present-from-to (channel-list-as-string channels-to-play) start-p end-p 'play)))
         (message "Now at the beginning of bar %d." (- start-bar-number 1)))))


; specialized function of process-selected-bars-laml-file-and-play
(defun first-number-of-string (str)
  (let ((p1 (find-in-string str 32)))
    (if p1 
        (string-to-number (substring str 0 p1))
        nil)))

(defun current-bar-number ()
  (save-excursion
    (find-next-bar-number)))

; specialized function of process-selected-bars-laml-file-and-play
(defun last-number-of-string (str)
  (let ((p1 (find-in-string str 32)))
    (if p1 
        (string-to-number (substring str p1 (length str)))
        nil)))

; (defun locate-initial-bar (bar-number)
;  (save-excursion
;    (goto-char (point-min))
;    (let ((pos (search-forward (concat "\"" (number-to-string bar-number) ":") nil t)))
;      (beginning-of-line 1)
;      (point))))
; 
; (defun locate-final-bar (bar-number)
;  (save-excursion
;    (goto-char (point-max))
;    (let ((pos (search-backward (concat "\"" (number-to-string bar-number) ":") nil t)))
;      (end-of-line 1)
;      (point))))

(defun locate-initial-bar (bar-number)
 "Return start point of bar-number"
 (save-excursion
   (beginning-of-line 1) (backward-up-list 1)  ; now at beginning of the current MidiTrack
   (let ((pos (search-forward (concat "\"" (number-to-string bar-number) ":") nil t)))
     (beginning-of-line 1)
     (point))))

(defun locate-final-bar (bar-number)
 "Return end point of bar-number"
 (save-excursion
   (beginning-of-line 1) (backward-up-list 1) (forward-sexp 1)  ; now at end of the current MidiTrack
   (let ((pos (search-backward (concat "\"" (number-to-string bar-number) ":") nil t)))
     (beginning-of-line 1) (forward-sexp 1)    ; if (end-of-line 1) is used it conflicts with a possible score comment
     (point))))

(defun process-selected-laml-file-certain-channels-and-play (channels)
  "Wrap the selected part of a midl file into a proper context, process it, and play it.
Works only for format 0 files."
  (interactive "sPlay which channels: (space-separated channel numbers): ")
  (let ((start-p (min (point) (mark)))
        (end-p (max (point) (mark))))
    (setq selection-begin start-p) (setq selection-end end-p) 
    (process-laml-file-and-present-from-to channels start-p end-p 'play)))

(defun process-selected-laml-file-certain-channels-and-play (channels)
  "Wrap the selected part of a midl file into a proper context, process it, and play it.
Works only for format 0 files."
  (interactive "sPlay which channels: (space-separated channel numbers): ")
  (let ((start-p (min (point) (mark)))
        (end-p (max (point) (mark))))
    (setq selection-begin start-p) (setq selection-end end-p) 
    (process-laml-file-and-present-from-to channels start-p end-p 'play)))

(defun process-selected-laml-file-certain-channels-and-illustrate (channels)
  "Wrap the selected part of a midl file into a proper context, process it, and illustrate it."
  (interactive "sPlay which channels: (space-separated channel numbers): ")
  (let ((start-p (min (point) (mark)))
        (end-p (max (point) (mark))))
    (setq selection-begin start-p) (setq selection-end end-p) 
    (process-laml-file-and-present-from-to channels start-p end-p 'illustrate)))

(defun process-bars-laml-file-certain-channels-and-play (channels start-end-bar-string)
  "Wrap certain chanenls of some bars of midl file into a proper context, process it, and play it."
  (interactive "sPlay which channels: (space-separated channel numbers): 
sFrom bar number to bar number (sep. by space), one bar number, or the current bar: ")
  (let* ((start-bar-number-0 (first-number-of-string start-end-bar-string))
         (start-bar-number (if start-bar-number-0 start-bar-number-0 (current-bar-number)))
         (end-bar-number (last-number-of-string start-end-bar-string))
         (start-p (locate-initial-bar start-bar-number))
         (end-p   (if end-bar-number (locate-final-bar end-bar-number) (locate-final-bar start-bar-number)))
        )
    (setq selection-begin start-p) (setq selection-end end-p) 
    (process-laml-file-and-present-from-to channels start-p end-p 'play)))

(defun process-bars-laml-file-certain-channels-and-illustrate (channels start-end-bar-string)
  "Wrap certain channels of some bars of midl file into a proper context, process it, and play it."
  (interactive "sPlay which channels: (space-separated channel numbers): 
sFrom bar number to bar number (sep. by space), one bar number, or the current bar: ")
  (let* ((start-bar-number-0 (first-number-of-string start-end-bar-string))
         (start-bar-number (if start-bar-number-0 start-bar-number-0 (current-bar-number)))
         (end-bar-number (last-number-of-string start-end-bar-string))
         (start-p (locate-initial-bar start-bar-number))
         (end-p   (if end-bar-number (locate-final-bar end-bar-number) (locate-final-bar start-bar-number)))
        )
    (setq selection-begin start-p) (setq selection-end end-p) 
    (process-laml-file-and-present-from-to channels start-p end-p 'illustrate)))


(defun midi-step (n)
  "Play the next n midi events, and advance point through these events to the beginning of the next midi event.
It is assumed that the points is at the start parenthesis of the first midi event. 
Leave mark at the beginning of the section."
  (interactive "p")
  (setq previous-midi-step-length n)
  (save-excursion
    (let ((p1 (point)))
       (forward-sexp n)
       (setq selection-begin p1) (setq selection-end (point)) 
       (process-laml-file-and-present-from-to t selection-begin selection-end 'play)
  ))
  (set-mark (point))
  (forward-sexp n)
  (while (looking-at-chars (point) '(9 10 13 32))   ; white space
         (forward-char 1)))

(defvar channels-to-play t "The boolean true (for all channels) or a list of integers in the interval [1..16].
  The single and barwise play operations take the value of this variable into account.")

(defun set-channels-to-play (channel-list-str)
  (interactive 
    (list (read-string (concat "Currently playing " (channel-info-string channels-to-play) ". Afterwards play "
               "(a space separated list of numbers or the symbol all): "))))
  (cond ((equal channel-list-str "all") 
           (setq channels-to-play t)
           (show-midl-mode-status-string) 
           (message "Thank you. All channels will play."))
        (t (setq channels-to-play (mapcar (function string-to-int) (string-to-list-separated-by 32 channel-list-str)))
           (show-midl-mode-status-string)
           (message (concat "Thank you. The following channels will play: " channel-list-str)))))

(defvar play-after-navigate t "A boolean that controls if we actually play the current note/bar after navigating to it.")

(defun toggle-play-after-navigate ()
  (interactive)
  (setq play-after-navigate (not play-after-navigate))
  (cond (play-after-navigate (message "Now play after navigating") )
        ((not play-after-navigate) (message "Now navigate only")))
  (show-midl-mode-status-string))

(defvar attribute-to-be-modified "note" "The name of the attribute to be modified by arrow left and arrow right")

(defvar possible-attribute-names (list "note" "velocity" "duration" "pressure" "value" "number") "All possible attribute names of all possible midi LAML form names")

; not used any more
(defun select-attribute-to-be-modified (attr-name-str)
 (interactive
   (list 
     (completing-read (concat "Current attribute to be modified: " 
                              attribute-to-be-modified ". "
                              "New attribute to be modified (space for possibilities): ")
                      '(("note" . nil) ("velocity" . nil) ("duration" . nil)))))
 (setq attribute-to-be-modified attr-name-str))


(defun toggle-attribute-to-be-modified ()
  (interactive)
  (let ((remaining (member attribute-to-be-modified possible-attribute-names)))
    (if remaining ; expect always true
        (let ((lgt-remaining (length remaining)))
          (cond ((= lgt-remaining 1)
                   (setq attribute-to-be-modified (car possible-attribute-names))
                   (show-midl-mode-status-string))
                ((> lgt-remaining 1)
                   (setq attribute-to-be-modified (cadr remaining))
                   (show-midl-mode-status-string))
                (t (error "toggle-attribute-to-be-modified: Should not happen"))))
        (error "toggle-attribute-to-be-modified: Should not happen")))
  (message "Now modifying attribute   %s" attribute-to-be-modified))
              
  

(defun string-to-list-separated-by (sep-char str)
  "Return the list of strings in str separated by one or more sep-char occurrences. 
str is not supposed to start with sep-char."
  (if (= (length str) 0)
      nil
      (filter (function (lambda (x) (not (empty-string-p x)))) 
        (string-to-list-separated-by-1 sep-char (eliminate-string-prefix (list sep-char) str)))))

(defun empty-string-p (x)
 (and (stringp x) (equal x "")))
 
(defun string-to-list-separated-by-1 (sep-char str)
  (let ((p (find-in-string str sep-char))
        (lgt (length str)))
    (if p
        (cons (substring str 0 p)
              (string-to-list-separated-by-1 sep-char (eliminate-string-prefix (list sep-char) (substring str p lgt))))
        (cons str nil))))

(defun eliminate-string-prefix (char-lst str)
  "Eliminate all prefix characters that belong to char-lst. 
Return the suffix list."
  (let ((i 0)
        (lgt (length str)))
    (while (and (< i lgt) (memq (aref str i) char-lst))
           (setq i (+ i 1)))
    (substring str i lgt)))

(defun channel-list-as-string (channel-list)
  "Transform channel-list (a list of integers) to a space separated string (without parentheses).
Pass the boolean input t through this function."
  (cond ((eq channel-list t) t)
        ((listp channel-list) (list-to-string (mapcar (function int-to-string) channel-list) " "))
        (t (error "channel-list-as-string: Expects either boolean t, or a list of integers."))))

(defun channel-info-string (channel-list)
  (cond ((eq channel-list t) "all channels")
        ((and (listp channel-list) (= (length channel-list) 1)) (concat "channel" " " (list-to-string (mapcar (function int-to-string) channel-list) " ")))
        ((and (listp channel-list) (> (length channel-list) 1)) (concat "channels" " " (list-to-string (mapcar (function int-to-string) channel-list) " ")))
        (t "???")))

(defun play-noteon (n)
  "Play the current noteOn midi message - and in total n noteOn midi events. Do not move point.
It is assumed that point is located at the start parenthesis of a NoteOn message, or just before it."
  (interactive "p")
  (save-excursion
    (let ((p1 (point)))
       (setq selection-begin p1)
       (if (eq channels-to-play t) 
           (forward-noteon n)
           (forward-noteon n channels-to-play))
       (message (concat "Playing a single NoteOn in " (channel-info-string channels-to-play)))
       (setq selection-end (point)) 
       (process-laml-file-and-present-from-to (channel-list-as-string channels-to-play) selection-begin selection-end 'play)))
)

(defun next-noteon-and-play ()
  "Go to next noteOn message and play it."
  (interactive)
  
  ; forwarding: Takes into consideration if point already is on a channel-relevant NoteOn message:
  (if (not (and (member (current-form-name) (list "NoteOn")) 
           (appropriate-current-channel channels-to-play)))
      (if (eq channels-to-play t) 
          (forward-noteon 0)
        (forward-noteon 0 channels-to-play))
      (if (eq channels-to-play t) 
          (forward-noteon 1)
        (forward-noteon 1 channels-to-play)))

  (if play-after-navigate
      (progn 
        (message (concat "Playing a single NoteOn in " (channel-info-string channels-to-play)))
        (if (not (looking-at ")"))      ; end of track
            (play-noteon 1)
          (progn (beep) (message "At end of track")))))
)

(defun previous-noteon-and-play ()
  "Go to previous noteOn message and play it."
  (interactive)

  ; backwarding: Takes into consideration if point already is on a channel-relevant NoteOn message:
  (if (not (and (member (current-form-name) (list "NoteOn")) 
           (appropriate-current-channel channels-to-play)))
      (if (eq channels-to-play t) 
          (backward-noteon 0) 
        (backward-noteon 0 channels-to-play))
      (if (eq channels-to-play t) 
          (backward-noteon 1) 
        (backward-noteon 1 channels-to-play)))
  (if play-after-navigate
     (progn 
       (message (concat "Playing a single NoteOn in " (channel-info-string channels-to-play)))
       (if (not (looking-at "MidiTrack")) ; beginning of track
           (play-noteon 1)
         (progn (beep) (message "At beginning of track")))))
)

; channels is a list of integers, or boolean t (all channels).
(defun forward-noteon (n &optional channels)
  "Move forward through n NoteOn midi messages. 
That is, pass over n channel-relevant NoteOn midi messages, and find the next channel-relevant NoteOn midi message.
If channels is provided (a list of channel integers) only move to a NoteOn that belongs to the given channels.
It is assumed that the point is located at the start parenthesis of a midi message, or just in front of it."
  (interactive "p")
  (let ((cnt n)
        (at-end-of-track nil)
       )
     (advance-through-white-space) (if (looking-at ")") (setq at-end-of-track t))

     ; Go past n relevant NoteOn messages:
     (while (and (not at-end-of-track) (> cnt 0))
       (if (and (member (current-form-name) (list "NoteOn")) 
                (appropriate-current-channel channels))  
           (setq cnt (- cnt 1)))
       (if (not at-end-of-track) (forward-sexp 1)) (advance-through-white-space)
       (if (looking-at ")") (setq at-end-of-track t))
     )

     ; move to next relevant NoteOn
     (while (and (not at-end-of-track) (not (and (member (current-form-name) (list "NoteOn")) 
                                                 (appropriate-current-channel channels))))
            (if (not at-end-of-track) (forward-sexp 1)) (advance-through-white-space)
            (if (looking-at ")") (setq at-end-of-track t)))))


; easier than forward-noteon, because (backward-sexp ...) lands on the start paren
; channels is a list of integers, or boolean t (all channels).
(defun backward-noteon (n &optional channels)
  "Move forward through n NoteOn midi messages. 
If channels is provided (a list of channel integers) only move to a NoteOn that belongs to the given channels.
It is assumed that the point is located at the start parenthesis of a midi message, or just in front of it."
  (interactive "p")
  (let ((cnt n)
        (at-beginning-of-track (looking-at "MidiTrack"))
       )

     (while (and (not at-beginning-of-track) (> cnt 0))
       (if (and (member (current-form-name) (list "NoteOn")) 
                (appropriate-current-channel channels))
           (setq cnt (- cnt 1)))
       (if (not at-beginning-of-track) (backward-sexp 1))
       (if (looking-at "MidiTrack") (setq at-beginning-of-track t))
     )

     (while (and (not at-beginning-of-track) (not (and (member (current-form-name) (list "NoteOn"))
                                                       (appropriate-current-channel channels))))
            (if (not at-beginning-of-track) (backward-sexp 1))
            (if (looking-at "MidiTrack") (setq at-beginning-of-track t)))

     ; move to previous NoteOn
))

(defun forward-midi-message (&optional n)
  "Navigate forward through the next midi  messages, and position point at initial parenthesis"
  (let ((n1 (if n n 1)))
    (forward-sexp n1)
    (advance-through-white-space)))

(defun backward-midi-message (&optional n)
  "Navigate backward through the next midi  messages, and position point at initial parenthesis"
  (let ((n1 (if n n 1)))
    (backward-sexp n1)))

; a helping function of forward-noteon and backward-noteon
; channels is a list of integers, or boolean t (all channels).
(defun appropriate-current-channel (channels)
  (let ((check-channels (if (null channels) nil t)))
    (if check-channels 
        (if (eq channels t)
            t
            (member (string-to-int (attribute-of-current-form 'channel)) channels))
        t)))
 
; ---------------------------------------------------------------------------------------------------
; Interactive attribute modification.

(defun do-add-to-attribute-of-current-form (level)
  (interactive)
  (let* ((attribute-name (string-to-symbol attribute-to-be-modified))
         (form-name (string-to-symbol (current-form-name)))
         (delta (get-attribute-delta-of-level level form-name attribute-name))
         (done (add-to-attribute-of-current-form attribute-name delta)))
    (if (and done play-after-navigate)
        (play-noteon 1))))

(defun do-subtract-from-attribute-of-current-form (level)
  (interactive)
  (let* ((attribute-name (string-to-symbol attribute-to-be-modified))
         (form-name (string-to-symbol (current-form-name)))
         (delta (get-attribute-delta-of-level level form-name attribute-name))
         (done (add-to-attribute-of-current-form attribute-name (- delta))))
    (if (and done play-after-navigate)
        (play-noteon 1))))



; Add delta to the value of attribute-name of current form.
; attribute-name is a symbol.
; Returns t if sucessful, else nil.
(defun add-to-attribute-of-current-form (attribute-name delta)
  (let* ((form-name (string-to-symbol (current-form-name)))
         (current-attribute-value-str (attribute-of-current-form attribute-name))
         (min-value (get-min-attribute-value form-name attribute-name))
         (max-value (get-max-attribute-value form-name attribute-name))
        )
    (if (or (not min-value) (not max-value))  ; invalid attribute-name
        (progn
           (beep)
           (message "Cannot set the attribute  %s  of a  %s  form" (symbol-to-string attribute-name) (current-form-name))
           nil)
        (if current-attribute-value-str
            (let* ((current-attribute-value (string-to-int current-attribute-value-str))
                   (new-value
                    (cond ((> delta 0) (min-maybe-unlimited max-value (+ delta current-attribute-value)))
                          ((< delta 0) (max-maybe-unlimited min-value (+ delta current-attribute-value)))
                          (t current-attribute-value))))
              (set-attribute-of-current-form attribute-name new-value)
              t)
            nil))))

(defun min-maybe-unlimited (t-or-number number)
  (if (eq t t-or-number) number (min t-or-number number)))

(defun max-maybe-unlimited (t-or-number number)
  (if (eq t t-or-number) number (max t-or-number number)))

; If nil is returned: attr-name invalid in form-name
; If t is returned: attr-name is unlimited minimum wise.
; If a number is returned: The number is the minimum.
(defun get-min-attribute-value (form-name attr-name)
  (cond ((eq form-name 'NoteOn)
           (cond ((eq attr-name 'note) 0)
                 ((eq attr-name 'velocity) 0)
                 ((eq attr-name 'duration) 0)
                 (t nil)))
        (t nil)))

; If nil is returned: attr-name invalid in form-name
; If t is returned: attr-name is unlimited minimum wise.
; If a number is returned: The number is the minimum.
(defun get-max-attribute-value (form-name attr-name)
  (cond ((eq form-name 'NoteOn)
           (cond ((eq attr-name 'note) 127)
                 ((eq attr-name 'velocity) 127)
                 ((eq attr-name 'duration) t)
                 (t nil)))
        (t nil)))

(defun get-attribute-delta-of-level (level form-name attribute-name)
  (cond ((and (eq form-name 'NoteOn) (eq attribute-name 'note))
           (cond ((= level 1) 1)
                 ((= level 2) 12)
                 ((= level 3) 24)
                 (t (error "get-delta-of-level: Unknown level of attribute %s of form %s" (symbol-to-string attribute-name) (symbol-to-string form-name)))))
        ((and (eq form-name 'NoteOn) (eq attribute-name 'velocity))
           (cond ((= level 1) 1)
                 ((= level 2) 10)
                 ((= level 3) 20)
                 (t (error "get-delta-of-level: Unknown level of attribute %s of form %s" (symbol-to-string attribute-name) (symbol-to-string form-name)))))
        ((and (eq form-name 'NoteOn) (eq attribute-name 'duration))
           (cond ((= level 1) 1)
                 ((= level 2) 100)
                 ((= level 3) 1000)
                 (t (error "get-delta-of-level: Unknown level of attribute %s of form %s" (symbol-to-string attribute-name) (symbol-to-string form-name)))))
        (t 1)))


; ---------------------------------------------------------------------------------------------------

(defun advance-through-white-space()   
  (while (or (looking-at-chars (point) '(9 10 13 32 59)) (is-inside-comment))   ; white space
         (forward-char 1)))

(defun next-similar-midi-step ()
  "Play the next previous-midi-step-length midi steps, continuing the previous midi step.
   Assume that the previous midi step left point at the beginning of a midi event."
  (interactive)
  (if previous-midi-step-length
      (progn 
        (save-excursion
          (let ((p1 (point)))
            (forward-sexp previous-midi-step-length)
            (setq selection-begin p1) (setq selection-end (point)) 
            (process-laml-file-and-present-from-to t p1 (point) 'play)))
        (set-mark (point))
        (forward-sexp previous-midi-step-length)
        (while (not (looking-at-chars (point) '(40))) ; start parenthesis
          (forward-char 1)))
      (message "Cannot repeat previous step.")))

(defun process-selected-laml-file-and-play-again ()
  "Play previous selection again"
  (interactive)
  (if last-temporary-midi-file-path
      (play-midi-file last-temporary-midi-file-path)  ; Alternatively, but slower: (process-laml-file-and-present-from-to selection-begin selection-end)
      (message "Cannot play again.")))

(defun process-laml-file-and-present-from-to (channels p1 p2 action)
  "Play or do other action from position p1 to p2. 
channels is a list of channel numbers (represented as a space separated text string).
action is either the symbol play or illustrate.
This function locates the relevant midi context and wraps it arround the selection."
  (interactive)
  (let* ((selected-message-txt (buffer-substring-no-properties p1 p2))
         (first-time-of-selection (find-first-abs-time-selection p1 p2))
        )
    (cond
           ((current-buffer-is-standard-midifile-of-format 0)
             (let* ((header (header-of-format-0-from-current-buffer))
                    (mode (mode-of-standard-midi-file-current-buffer))
                    (tempo-context (find-global-tempo-event))
                    (context (cond ((eq mode 'absTime)
                                      (find-local-voice-context-abs-time-and-reset-to 1920))  ; earlier: (context-of-format-0-from-current-buffer)
                                   ((eq mode 'deltaTime) 
                                      (find-local-voice-context-delta-time))
                                   (t (error "process-laml-file-and-present-from-to: Unknown mode:" mode))))
                    (last-time-of-context 1920)   ; earlier: (find-last-abs-time-of context)
                    (temp-file-path (concat (current-directory) "midi-temp/" "sel-temp.midl"))
                    (midi-file-path (concat (current-directory) "midi-temp/" "midi-temp/" "sel-temp.mid"))
                   )
              (setq last-temporary-midi-file-path midi-file-path)
              (ensure-directory-existence (current-directory) "midi-temp/")
              (ensure-directory-existence (concat (current-directory) "midi-temp/") "midi-temp/")
              (if (file-exists-p temp-file-path) (delete-file temp-file-path))
              (cond ((and (eq mode 'absTime) first-time-of-selection last-time-of-context)
                       (let ((buf (make-a-file-from-laml-template 
                                   "sel-temp.midl"
                                   (concat (current-directory) "midi-temp/")
                                   "midi-file-for-selection-time-displace"
                                   'midi-laml-mode
                                   (list
                                    (list "THE-HEADER" header)
                                    (list "THE-CONTEXT" (concat tempo-context CR context))
                                    (list "THE-SELECTION" selected-message-txt)
                                    (list "THE-CHANNELS" (if (and (symbolp channels) (eq channels t)) "#t" (concat "'(" channels ")")))
                                    (list "THE-TIME" (int-to-string (- (- (- first-time-of-selection last-time-of-context) 1)))) ; negative time displacement...
                                    )
                                   (concat laml-dir "styles/xml-in-laml/midi/" "templates/"))))
                          (kill-buffer buf)))

                     ((eq mode 'absTime)
                        (error "Could, of some reason not find the last time of context or the first time of selection"))

                     ((and (eq mode 'deltaTime))
                       (let ((buf (make-a-file-from-laml-template 
                                   "sel-temp.midl"
                                   (concat (current-directory) "midi-temp/")
                                   "midi-file-for-selection-no-time-displacement"
                                   'midi-laml-mode
                                   (list
                                    (list "THE-HEADER" header)
                                    (list "THE-CONTEXT" context)
                                    (list "THE-CHANNELS" (if (and (symbolp channels) (eq channels t)) "#t" (concat "'(" channels ")")))
                                    (list "THE-SELECTION" selected-message-txt)
                                    )
                                   (concat laml-dir "styles/xml-in-laml/midi/" "templates/"))))
                          (kill-buffer buf)))

                     (t (error "Not made yet")))
               (copy-file path-to-midi-error-file midi-file-path t)   ; error midi file copying
               (laml-sync-process-file temp-file-path 'silent)
;               (message (concat "A MIDI file has been generated in midi-temp. " (if (eq action 'play) "Now playing." "Now illustrating.")))
               (cond ((file-exists-p midi-file-path) ; with possible double name
                        (cond ((eq action 'play) (play-midi-file midi-file-path))
                              ((eq action 'illustrate) (illustrate-midi-file midi-file-path))
                              (t (error "process-laml-file-and-present-from-to: Unknown action"))))
                     (t (error "Process-laml-file-and-play: Should not happen:")))))

          ((current-buffer-is-standard-midifile-of-format 1)
             (let* ((header (header-of-format-1-from-current-buffer))
                    (mode (mode-of-standard-midi-file-current-buffer))
                    (context-1 "")          ; earlier  (global-context-of-format-1-from-current-buffer-first-bar)
                    (context-2 (cond ((eq mode 'absTime)
                                      (find-local-voice-context-abs-time-and-reset-to 1920))  ; earlier: (context-of-format-0-from-current-buffer)
                                   ((eq mode 'deltaTime) 
                                      (find-local-voice-context-delta-time))
                                   (t (error "process-laml-file-and-present-from-to: Unknown mode:" mode))))
                    (tempo-context (find-global-tempo-event))
                    (last-time-of-context 1920) ; earlier  (find-last-abs-time-of context-2)
                    (temp-file-path (concat (current-directory) "midi-temp/" "sel-temp.midl"))
                    (midi-file-path (concat (current-directory) "midi-temp/" "midi-temp/" "sel-temp.mid"))
                   )
              (setq last-temporary-midi-file-path midi-file-path)
              (ensure-directory-existence (current-directory) "midi-temp/")
              (ensure-directory-existence (concat (current-directory) "midi-temp/") "midi-temp/")
              (if (file-exists-p temp-file-path) (delete-file temp-file-path))
              (cond ((and (eq mode 'absTime) first-time-of-selection last-time-of-context)
                       (let ((buf (make-a-file-from-laml-template 
                                   "sel-temp.midl"
                                   (concat (current-directory) "midi-temp/")
                                   "midi-file-for-selection-time-displace"
                                   'midi-laml-mode
                                   (list
                                    (list "THE-HEADER" header)
                                    (list "THE-CONTEXT" (concat tempo-context CR context-1 CR context-2))
                                    (list "THE-SELECTION" selected-message-txt)
                                    (list "THE-CHANNELS" (if (and (symbolp channels) (eq channels t)) "#t" (concat "'(" channels ")")))
                                    (list "THE-TIME" (int-to-string (- (- (- first-time-of-selection last-time-of-context) 1)))) ; negative time displacement...
                                    )
                                   (concat laml-dir "styles/xml-in-laml/midi/" "templates/"))))
                          (kill-buffer buf)))

                     ((eq mode 'absTime)
                        (error "Could, of some reason not find the last time of context or the first time of selection"))

                     ((and (eq mode 'deltaTime))             ; merge track0 with selected part (and its local context):
                       (let ((buf (make-a-file-from-laml-template 
                                   "sel-temp.midl"
                                   (concat (current-directory) "midi-temp/")
                                   "midi-file-for-selection-no-time-displacement-merge-contexts"
                                   'midi-laml-mode
                                   (list
                                    (list "THE-HEADER" header)
                                    (list "THE-FIRST-CONTEXT" context-1)
                                    (list "THE-SECOND-CONTEXT" context-2)
                                    (list "THE-CHANNELS" (if (and (symbolp channels) (eq channels t)) "#t" (concat "'(" channels ")")))
                                    (list "THE-SELECTION" selected-message-txt)
                                    )
                                   (concat laml-dir "styles/xml-in-laml/midi/" "templates/"))))
                          (kill-buffer buf)))

                     (t (error "Not made yet")))
               (copy-file path-to-midi-error-file midi-file-path t)   ; error midi file copying
               (laml-sync-process-file temp-file-path 'silent)
;               (message (concat "A MIDI file has been generated in midi-temp. " (if (eq action 'play) "Now playing." "Now illustrating.")))
               (cond ((file-exists-p midi-file-path) ; with possible double name
                        (cond ((eq action 'play) (play-midi-file midi-file-path))
                              ((eq action 'illustrate) (illustrate-midi-file midi-file-path))
                              (t (error "process-laml-file-and-present-from-to: Unknown action"))))
                     (t (error "Process-laml-file-and-play: Should not happen:")))))

          ((current-buffer-is-multipad-format)
             (let* ((header (header-of-multipad-from-current-buffer))
                    (mode (mode-of-multipad-current-buffer))
                    (context (cond ((eq mode 'absTime)
                                      (find-local-voice-context-abs-time-and-reset-to 1920))  ; earlier: (context-of-multipad-from-current-buffer)
                                   ((eq mode 'deltaTime) 
                                      (find-local-voice-context-delta-time))
                                   (t (error "process-laml-file-and-present-from-to: Unknown mode:" mode))))
                    (last-time-of-context 1)  ; earlier: (find-last-abs-time-of context)
                    (temp-file-path (concat (current-directory) "midi-temp/" "sel-temp.midl"))
                    (midi-file-path (concat (current-directory) "midi-temp/" "midi-temp/" "sel-temp.mid"))
                   )
              (setq last-temporary-midi-file-path midi-file-path)
              (ensure-directory-existence (current-directory) "midi-temp/")
              (ensure-directory-existence (concat (current-directory) "midi-temp/") "midi-temp/")
              (if (file-exists-p temp-file-path) (delete-file temp-file-path))
              (cond ((and (eq mode 'absTime) first-time-of-selection last-time-of-context)
                       (let ((buf (make-a-file-from-laml-template 
                                   "sel-temp.midl"
                                   (concat (current-directory) "midi-temp/")
                                   "midi-file-for-selection-time-displace"
                                   'midi-laml-mode
                                   (list
                                    (list "THE-HEADER" header)
                                    (list "THE-CONTEXT" context)
                                    (list "THE-SELECTION" selected-message-txt)
                                    (list "THE-CHANNELS" (if (and (symbolp channels) (eq channels t)) "#t" (concat "'(" channels ")")))
                                    (list "THE-TIME" (int-to-string (- (- (- first-time-of-selection last-time-of-context) 1)))) ; negative time displacement...
                                    )
                                   (concat laml-dir "styles/xml-in-laml/midi/" "templates/"))))
                          (kill-buffer buf)))

                     ((eq mode 'absTime)
                        (error "Could, of some reason not find the last time of context or the first time of selection"))

                     ((and (eq mode 'deltaTime))   
                       (let ((buf (make-a-file-from-laml-template 
                                   "sel-temp.midl"
                                   (concat (current-directory) "midi-temp/")
                                   "midi-file-for-selection-no-time-displacement"
                                   'midi-laml-mode
                                   (list
                                    (list "THE-HEADER" header)
                                    (list "THE-CONTEXT" context)
                                    (list "THE-CHANNELS" (if (and (symbolp channels) (eq channels t)) "#t" (concat "'(" channels ")")))
                                    (list "THE-SELECTION" selected-message-txt)
                                    )
                                   (concat laml-dir "styles/xml-in-laml/midi/" "templates/"))))
                          (kill-buffer buf)))

                     (t (error "Not made yet")))
               (copy-file path-to-midi-error-file midi-file-path t)   ; error midi file copying
               (laml-sync-process-file temp-file-path 'silent)
;               (message (concat "A MIDI file has been generated in midi-temp. " (if (eq action 'play) "Now playing." "Now illustrating.")))
               (cond ((file-exists-p midi-file-path) ; with possible double name
                      (cond ((eq action 'play) (play-midi-file midi-file-path))
                            ((eq action 'illustrate) (illustrate-midi-file midi-file-path))
                            (t (error "process-laml-file-and-present-from-to: Unknown action"))))
                     (t (error "Process-laml-file-and-play: Should not happen:")))))

          ((current-buffer-is-style-format)
             (let* ((header (header-of-style-from-current-buffer))
                    (mode (mode-of-style-current-buffer))
                    (context (cond ((eq mode 'absTime)
                                      (find-context-in-style-setup-section))
                                   ((eq mode 'deltaTime) 
                                      (error "As of now, TyrosStyle only supports absTime mode"))
                                   (t (error "process-laml-file-and-present-from-to: Unknown mode:" mode))))
                    (last-time-of-context 0)
                    (temp-file-path (concat (current-directory) "midi-temp/" "sel-temp.midl"))
                    (midi-file-path (concat (current-directory) "midi-temp/" "midi-temp/" "sel-temp.mid"))
                   )
              (setq last-temporary-midi-file-path midi-file-path)
              (ensure-directory-existence (current-directory) "midi-temp/")
              (ensure-directory-existence (concat (current-directory) "midi-temp/") "midi-temp/")
              (if (file-exists-p temp-file-path) (delete-file temp-file-path))
              (cond ((and (eq mode 'absTime) first-time-of-selection last-time-of-context)
                       (let ((buf (make-a-file-from-laml-template 
                                   "sel-temp.midl"
                                   (concat (current-directory) "midi-temp/")
                                   "midi-file-for-selection-time-displace"
                                   'midi-laml-mode
                                   (list
                                    (list "THE-HEADER" header)
                                    (list "THE-CONTEXT" context)
                                    (list "THE-SELECTION" selected-message-txt)
                                    (list "THE-CHANNELS" (if (and (symbolp channels) (eq channels t)) "#t" (concat "'(" channels ")")))
                                    (list "THE-TIME" (int-to-string (- (- (- first-time-of-selection last-time-of-context) 1)))) ; negative time displacement...
                                    )
                                   (concat laml-dir "styles/xml-in-laml/midi/" "templates/"))))
                          (kill-buffer buf)))

                     ((eq mode 'absTime)
                        (error "Could, of some reason not find the last time of context or the first time of selection"))

                     ((and (eq mode 'deltaTime))  ; no yet relevant - but kept because it may be useful in the future
                       (let ((buf (make-a-file-from-laml-template 
                                   "sel-temp.midl"
                                   (concat (current-directory) "midi-temp/")
                                   "midi-file-for-selection-no-time-displacement"
                                   'midi-laml-mode
                                   (list
                                    (list "THE-HEADER" header)
                                    (list "THE-CONTEXT" context)
                                    (list "THE-CHANNELS" (if (and (symbolp channels) (eq channels t)) "#t" (concat "'(" channels ")")))
                                    (list "THE-SELECTION" selected-message-txt)
                                    )
                                   (concat laml-dir "styles/xml-in-laml/midi/" "templates/"))))
                          (kill-buffer buf)))

                     (t (error "Not made yet")))
               (copy-file path-to-midi-error-file midi-file-path t)   ; error midi file copying
               (laml-sync-process-file temp-file-path 'silent)
;               (message (concat "A MIDI file has been generated in midi-temp. " (if (eq action 'play) "Now playing." "Now illustrating.")))
               (cond ((file-exists-p midi-file-path) ; with possible double name
                      (cond ((eq action 'play) (play-midi-file midi-file-path))
                            ((eq action 'illustrate) (illustrate-midi-file midi-file-path))
                            (t (error "process-laml-file-and-present-from-to: Unknown action"))))
                     (t (error "Process-laml-file-and-play: Should not happen:")))))

          (t (error "Can only select, process and play a standard midi file of format 0 or 1, a Tyros Multipad, or a Tyros Style")))))


(defun find-global-tempo-event()
 (save-excursion
   (goto-char (point-min))
   (let ((search-res (search-forward " Tempo: " nil t)))
     (if search-res
         (progn
            (beginning-of-line 1)
            (let ((p0 (point)))
               (forward-sexp 1)
               (buffer-substring-no-properties p0 (point))))
         ""))))



(defun find-first-abs-time-selection (p1 p2)
  (save-excursion
    (goto-char p1)
    (forward-sexp 1)
    (let ((pn (point)))
      (goto-char p1)
      (if (search-forward "'absTime" pn t)
          (progn
             (search-forward "\"")
             (string-to-int (prog-name-under-point)))
          nil))))

(defun find-first-delta-time-selection (p1 p2)
  (save-excursion
    (goto-char p1)
    (forward-sexp 1)
    (let ((pn (point)))
      (goto-char p1)
      (if (search-forward "'deltaTime" pn t)
          (progn
             (search-forward "\"")
             (string-to-int (prog-name-under-point)))
          nil))))             

; To general  
(defun ensure-directory-existence (prefix-dir dir)
 "Ensure that the directory with path (concat prefix-dir dir) exists.
If necessary, create dir in prefix-dir. It is assumed as a precondition that
prefix-dir exists before the activation of this procedure."
  (if (not (file-exists-p (concat prefix-dir dir))) 
      (make-directory (concat prefix-dir dir))))

(defun current-buffer-is-standard-midifile ()
  "Is the current buffer a standard midi file (as opposed to a tyros multipad)."
  (save-excursion
    (goto-char (point-min))
    (if (and (search-forward "(StandardMidiFile" nil t)
             (search-forward "(MidiHeader" nil t))
        t
        nil)))

; Is the current buffer a format f standard midi file.
; f is an integer number, such as 0, 1 or 2.
; It is assumed that the format is given as a text string of the format attribute in the MidiHeader.
; A TyrosMultiPad is not, in this context, a standard midi file.
(defun current-buffer-is-standard-midifile-of-format (f)
  (save-excursion
    (goto-char (point-min))
    (if (and (search-forward "(StandardMidiFile" nil t)
             (search-forward "(MidiHeader" nil t))
        (progn
          (backward-up-list 1)
          (let ((p1 (point))
                (p2 (progn (forward-sexp 1) (point))))
            (backward-sexp 1)
            (if (and (search-forward "'format" p2 t) (search-forward "\""))
                (let ((format-char (char-under-point)))
                  (= format-char (+ ?0 f)))
                nil)))
        nil)))

(defun header-of-format-0-from-current-buffer ()
  (save-excursion
    (goto-char (point-min))
    (if (and (search-forward "(StandardMidiFile" nil t)
             (search-forward "(MidiHeader" nil t))
        (progn
          (backward-up-list 1)
          (let ((p0 (point)))
             (forward-sexp 1)
             (buffer-substring-no-properties p0 (point))))
        nil)))

(defun context-of-format-0-from-current-buffer ()
  (save-excursion
    (goto-char (point-min))
    (if (and (search-forward "(StandardMidiFile" nil t)
             (search-forward "(MidiTrack" nil t))
      (let ((p0 (point)))
        (search-forward "(NoteOn" nil t)
        (backward-up-list 1)
        (buffer-substring-no-properties p0 (point)))
      nil)))




(defun find-last-abs-time-of (fragment)
  "Return the abs-time of the last midi event in fragment (a string). Returns an integer, or nil."
  (let ((search-res (string-search-backward fragment "'absTime")))
    (if search-res
        (let* ((p1 (find-in-string fragment 34 search-res))    ; 34 = '"'
               (p2 (find-in-string fragment 34 (+ p1 1))))
           (string-to-int (substring fragment (+ p1 1) p2)))
        nil)))

  
(defun mode-of-standard-midi-file-current-buffer  ()
  "Return the mode (absTime, deltaTime) of the current buffer, which
as a precondition is assumed to hold a StandardMidiFile of arbitrary format (0, 1)."
  (save-excursion
    (goto-char (point-min))
    (if (and (search-forward "(StandardMidiFile" nil t)
             (search-forward "(MidiHeader" nil t))
        (progn
          (backward-up-list 1)
          (let ((p0 (point)))
             (forward-sexp 1)
             (let ((p1 (point)))
                (goto-char p0)
                (if (search-forward "'mode" p1 t)
                    (progn
                       (search-forward "\"")
                       (string-to-symbol (prog-name-under-point)))
                    nil))))
         nil)))

; --------------------------------------------------------------------------------------------------
; Format 1 stuff.

; Return a midi file format 0 header corresponding to the format 1 melody of the current buffer.
; A string.
(defun header-of-format-1-from-current-buffer ()
 (let ((mode-string (symbol-to-string (mode-of-standard-midi-file-current-buffer)))
       (ppqn-str (number-to-string (find-pulses-per-quarter-note-of-current-buffer)))
      )
  (concat 
    "(" "MidiHeader" " " "'format" " " "\"0\"" " " "'numberOfTracks" " " "\"1\"" " "
        "'pulsesPerQuarterNote" " " (string-it ppqn-str) " "
        "'mode" " " (string-it mode-string) ")"
        )) )

; Not used
(defun context-of-format-1-from-current-buffer ()
  (concat 
    (global-context-of-format-1-from-current-buffer-first-bar) CR
    (local-context-of-format-1-from-current-buffer-first-bar)
  )
)

; Return the entire first track
(defun global-context-of-format-1-from-current-buffer ()
  (save-excursion
    (goto-char (point-min))
    (if (and (search-forward "(StandardMidiFile" nil t)
             (search-forward "(MidiTrack" nil t))
      (let ((p0 (point)))
        (backward-up-list 1) 
        (forward-sexp 1)  ; through the entire MidiTrack
        (backward-char 1)
        (forward-sexp -1) ; back over end of track meta event
        (buffer-substring-no-properties p0 (- (point) 1)))
      nil)))




; If possible, return the first bar of first track. 
; If not possible, return entire contents of first track
(defun global-context-of-format-1-from-current-buffer-first-bar ()
  (save-excursion
    (goto-char (point-min))
    (if (and (search-forward "(StandardMidiFile" nil t)
             (search-forward "(MidiTrack" nil t))
        (let ((p0 (point)))
          (let ((res (search-forward "\"1:"))) ; start of bar 1
            (if res
                (progn
                  (beginning-of-line)
                  (buffer-substring-no-properties p0 (point)))
              (global-context-of-format-1-from-current-buffer))))
        "")))


(defun local-context-of-format-1-from-current-buffer ()
  (save-excursion
    (if (search-backward "(MidiTrack" nil t)
      (progn
         (search-forward "(MidiTrack" nil t)
         (let ((p0 (point)))
           (search-forward "(NoteOn" nil t)
           (backward-up-list 1)
           (buffer-substring-no-properties p0 (point))))
      "")) )


; (defun find-local-voice-context-abs-time-and-reset-to (new-time)
;   "Find and return the closest ProgramChange and corresponding ControlChange context in the current MidiTrack.
; Reset the abs-time to new-time (an integer).
; Operates relative to point.
; Returns a relatively brief string. Returns an empty string if an appropriate context cannot be found." 
;   (save-excursion
;     (if (search-backward "(MidiTrack" nil t)
;         (let ((track-start (point)) (track-end nil) (p1 nil) (p2 nil) (p-program-change nil))
;           (forward-sexp 1) 
;           (setq track-end (point))   ; end of track position
;           (backward-sexp 1)      ; now at start of MidiTrack
;           (search-forward "(MidiTrack" nil t)
;           
;           (if (search-forward "(ProgramChange" track-end t)
;               (progn
;                 (search-backward "(ProgramChange" nil t)
;                 (setq p-program-change (point))
;                 (let ((abs-time-program-change (abs-time-current-form)))
;                   (while (and (member (current-form-name) (list "ControlChange" "ProgramChange")) (= abs-time-program-change (abs-time-current-form)))
;                     (setq p1 (point))
;                     (backward-sexp 1))
;                   (goto-char p-program-change)
;                   (while (and (member (current-form-name) (list "ControlChange" "ProgramChange")) (= abs-time-program-change (abs-time-current-form)))
;                     (forward-sexp 1)
;                     (setq p2 (point))
;                     (while (not (looking-at "(")) (forward-char 1)))
;                   (reset-context-from-to-time (buffer-substring-no-properties p1 p2) abs-time-program-change new-time)))
;             "")
;           )
;           "")))

(defun find-local-voice-context-abs-time-and-reset-to (new-time)
  "Find and return the closest ProgramChange and corresponding ControlChange context in the current MidiTrack.
Reset the abs-time to new-time (an integer).
Operates relative to point.
Returns a relatively brief string. Returns an empty string if an appropriate context cannot be found." 
  (save-excursion
    (let ((p-cur (point)))
      (if (search-backward "(MidiTrack" nil t)
          (let ((track-start (point)) (track-end nil) (p1 nil) (p2 nil) (p-program-change nil))
            (forward-sexp 1) 
            (setq track-end (point))    ; end of track position
            (goto-char p-cur)
          
            (if (search-backward "(ProgramChange" track-start t)
                (progn
                  (setq p-program-change (point))
                  (let ((abs-time-program-change (abs-time-current-form)))
                    (while (and (member (current-form-name) (list "ControlChange" "ProgramChange")) (= abs-time-program-change (abs-time-current-form)))
                      (setq p1 (point))
                      (backward-sexp 1))
                    (goto-char p-program-change)
                    (while (and (member (current-form-name) (list "ControlChange" "ProgramChange")) (= abs-time-program-change (abs-time-current-form)))
                      (forward-sexp 1)
                      (setq p2 (point))
                      (while (not (looking-at "(")) (forward-char 1)))
                    (reset-context-from-to-time (buffer-substring-no-properties p1 p2) abs-time-program-change new-time)))
              "")
            )
        ""))))


(defun find-local-voice-context-delta-time ()
  "Find and return the closest ProgramChange and corresponding ControlChange context in the current MidiTrack.
Operates relative to point.
Returns a relatively brief string. Returns an empty string if an appropriate context cannot be found." 
  (save-excursion
    (if (search-backward "(MidiTrack" nil t)
        (let ((track-start (point)) (track-end nil) (p1 nil) (p2 nil) (p-program-change nil))
          (forward-sexp 1) 
          (setq track-end (point))   ; end of track position
          (backward-sexp 1)      ; now at start of MidiTrack
          (search-forward "(MidiTrack" nil t)
          
          (if (search-forward "(ProgramChange" track-end t)
              (progn
                (search-backward "(ProgramChange" nil t)
                (setq p-program-change (point))
                (let ((delta-time-program-change (delta-time-current-form)))
                  (while (and (member (current-form-name) (list "ControlChange" "ProgramChange")) (= delta-time-program-change (delta-time-current-form)))
                    (setq p1 (point))
                    (backward-sexp 1))
                  (goto-char p-program-change)
                  (while (and (member (current-form-name) (list "ControlChange" "ProgramChange")) (= delta-time-program-change (delta-time-current-form)))
                    (forward-sexp 1)
                    (setq p2 (point))
                    (while (not (looking-at "(")) (forward-char 1)))
                  (buffer-substring-no-properties p1 p2)))
            "")
          )
          ""))
)  

(defun abs-time-current-form ()
  "Return the absTime of the current form. Point must be a initial parenthesis."
  (save-excursion
     (let ((p1 (point)))
       (forward-sexp 1)
       (let ((p2 (point)))
          (find-first-abs-time-selection p1 p2)))))

(defun delta-time-current-form ()
  "Return the deltaTime of the current form. Point must be a initial parenthesis."
  (save-excursion
     (let ((p1 (point)))
       (forward-sexp 1)
       (let ((p2 (point)))
          (find-first-delta-time-selection p1 p2)))))


(defun current-form-name()
  "Return the name of the current form. Point must be pointing at the initial parenthesis, or just in front of it."
  (save-excursion
    (while (or (looking-at "(") (looking-at " ")) (forward-char 1))
    (prog-name-under-point)))

(defun attribute-of-current-form (attribute-name)
   "Return the attribute named attribute-name of the current form (a string).
Return nil if attribute is not found.
Point is assumed to be at - or just before the initial parenthesis of the form." 
  (save-excursion
    (while (looking-at-chars (point) white-space-char-list) (forward-char 1))
    (let ((p0 (point)))
      (end-of-line 1) 
      (let ((p1 (point)))
         (ast-list-attribute (car (read-from-string (buffer-substring-no-properties p0 p1))) attribute-name)))))


(defun set-attribute-of-current-form (attr-name new-value)
  "In the current form, set the value of the attribute named attr-name (a symbol) to new-value (an integer).
A fatal error occurs if attr-name is not present in the current form."
  (let ((attr-name-str (symbol-to-string attr-name))
        (new-value-str (int-to-string new-value)))
   (save-excursion
     (let ((p0 (point)))
       (forward-sexp 1)
       (let ((p1 (point)))
         (backward-sexp 1)
         (let ((search-res (search-forward (concat "'" attr-name-str) p1 t)))
           (if search-res
               (progn
                 (while (not (looking-at "\"")) (forward-char 1)) (forward-char 1) ; now point is at first ciffer in attr value
                 (while (not (looking-at "\"")) (delete-char 1)) ; delete existing attribute value
                 (insert new-value-str))
               (error "Cannot find and change the value of attribute %s to %d in form %s." attr-name-str new-value (current-form-name)))))))))
                
  

(defun reset-context-from-to-time (context-string from-time to-time)
  "Be sure that all events in context-string starts at time time (an integer). Return the reset context-string"
   (let* ((temp-buf-name "*replace-buffer*")
          (temp-buf (if (get-buffer temp-buf-name)
                        (get-buffer temp-buf-name)
                        (generate-new-buffer temp-buf-name)))
          (cur-buffer (current-buffer))
          (from-time-string (if (numberp from-time) (int-to-string from-time) from-time))
          (to-time-string (if (numberp to-time) (int-to-string to-time) to-time))
          (abstime-clause-reg-exp
             (regular-expression  `(concat "'absTime" (one-or-more (char-set tab newline space return)) "\"" ,(number-to-string from-time) "\""))
          )
         )
    (set-buffer temp-buf)
    (erase-buffer)
    (insert context-string) (goto-char (point-min))
    (while (re-search-forward abstime-clause-reg-exp nil t)
       (replace-match  (concat "'absTime" " "  (concat "\"" to-time-string "\"")) nil nil))
    (let ((res (buffer-string)))
       (set-buffer cur-buffer)
       res)))

(defun local-context-of-format-1-from-current-buffer-first-bar ()
  (save-excursion
    (if (search-backward "(MidiTrack" nil t)
      (progn
         (search-forward "(MidiTrack" nil t)
         (let ((p0 (point)))
           (let ((res (search-forward "\"1:"))) ; start of bar 1
             (if res
                 (progn
                   (beginning-of-line)
                   (buffer-substring-no-properties p0 (point)))
               (local-context-of-format-1-from-current-buffer)))))
       "")))





; ---------------------------------------------------------------------------------------------------
; Multipad stuff.

(defun current-buffer-is-multipad-format ()
  (save-excursion
    (goto-char (point-min))
    (if (and (search-forward "(TyrosMultiPad" nil t)
             (search-forward "(MidiTrack" nil t))
        t
        nil)))

; Relative to the current position, find and return the context
(defun context-of-multipad-from-current-buffer ()
  (save-excursion
    (if (search-backward "(MidiTrack" nil t)
      (progn
         (search-forward "(MidiTrack" nil t)
         (let ((p0 (point)))
           (search-forward "(NoteOn" nil t)
           (backward-up-list 1)
           (buffer-substring-no-properties p0 (point))))
      nil)) )

(defun header-of-multipad-from-current-buffer ()
 "Return a midi file format 0 header corresponding to the multi pad of the current buffer. A string."
 (let ((mode-string (symbol-to-string (mode-of-multipad-current-buffer))))
  (concat 
    "(" "MidiHeader" " " "'format" " " "\"0\"" " " "'numberOfTracks" " " "\"1\"" " "
        "'pulsesPerQuarterNote" " " "\"1920\"" " "
        "'mode" " " (string-it mode-string) ")"
        )))

(defun mode-of-multipad-current-buffer ()
  "Returns the mode of the multipad buffer. A symbol."
  (save-excursion
    (goto-char (point-min))
    (if (search-forward "(TyrosMultiPad" nil t)
        (progn
          (backward-up-list 1)
          (let ((p0 (point)))
             (forward-sexp 1)
             (let ((p1 (point)))
                (goto-char p0)
                (if (search-forward "'mode" p1 t)
                    (progn
                       (search-forward "\"")
                       (string-to-symbol (prog-name-under-point)))
                    nil))))
         nil)) )

; ---------------------------------------------------------------------------------------------------------------
; Tyros style stuff.

(defun current-buffer-is-style-format ()
  (save-excursion
    (goto-char (point-min))
    (if (and (search-forward "(TyrosStyle" nil t)
             (search-forward "(MidiSection" nil t))
        t
        nil)))

; Returns a symbol.
(defun mode-of-style-current-buffer ()
  (save-excursion
    (goto-char (point-min))
    (if (search-forward "(TyrosStyle" nil t)
        (progn
          (backward-up-list 1)
          (let ((p0 (point)))
             (forward-sexp 1)
             (let ((p1 (point)))
                (goto-char p0)
                (if (search-forward "'mode" p1 t)
                    (progn
                       (search-forward "\"")
                       (string-to-symbol (prog-name-under-point)))
                    nil))))
         nil)) )

(defun find-context-in-style-setup-section ()
  (save-excursion
    (goto-char (point-min))
    (if (search-forward "(TyrosStyle" nil t)
        (if (search-forward "(MidiSection" nil t)
            (let ((p0 (search-forward "(StyleSetup" nil t)))
               (if p0
                   (progn
                     (backward-up-list 1)
                     (forward-sexp 1)
                     (let ((pn (point)))
                       (goto-char p0)
                       (forward-through-white-space)
                       (buffer-substring-no-properties (point) (- pn 1))))
                   ""))
             "")
         "")))
        
(defun header-of-style-from-current-buffer ()
 "Return a midi file format 0 header corresponding to the style of the current buffer. A string."
 (let ((mode-string (symbol-to-string (mode-of-style-current-buffer))))
  (concat 
    "(" "MidiHeader" " " "'format" " " "\"0\"" " " "'numberOfTracks" " " "\"1\"" " "
        "'pulsesPerQuarterNote" " " "\"1920\"" " "
        "'mode" " " (string-it mode-string) ")"
        )))

; ---------------------------------------------------------------------------------------------------------------
; Midi list length function

(defun display-length-of-midi-list ()
  "Displays information about the length of the selected midi list (an integer) in the mini buffer.
Works in both (clean and exclusive) absTime and deltaTime modes.
Does not include the first deltaTime if applied on a selection of deltaTime midi messages."
  (interactive)
  (let* ((first-lgt (length-of-first-message-of-current-buffer))
         (lgt (length-of-midi-list))
         (mode (guess-processing-mode  (min (point) (mark)) (max (point) (mark))))  ; a string: "absTime" or "deltaTime"   
         (ppqn (find-pulses-per-quarter-note-of-current-buffer))  ; an integer
         (number-of-quater-notes (/ lgt ppqn))
         (rest-lgt (mod lgt ppqn))
        )
    (cond 
          ((and (equal mode "absTime") (>= lgt ppqn))
             (message "The length is %d pulses = %d quarter notes + %d pulses (ppqn=%d)." lgt number-of-quater-notes rest-lgt ppqn))

          ((equal mode "absTime")
             (message "The length is %d pulses." lgt))

          ((and (equal mode "deltaTime") (>= lgt ppqn))
             (message "The length is %d pulses = %d quarter notes + %d pulses (ppqn=%d).  If first event is included %d pulses = %d quarter notes + %d pulses."
                      lgt number-of-quater-notes rest-lgt  ppqn (+ lgt first-lgt) (/ (+ lgt first-lgt) ppqn) (mod (+ lgt first-lgt) ppqn)  ))

          ((and (equal mode "deltaTime") (>= (+ lgt first-lgt) ppqn))
             (message "The length is %d pulses.  If first event is included %d quarter notes + %d pulses (ppqn=%d)."
                      lgt (/ (+ lgt first-lgt) ppqn) (mod (+ lgt first-lgt) ppqn) ppqn  ))

          ((equal mode "deltaTime")
             (message "The length is %d pulses: ppqn=%d.  If first event is include: %d pulses."
                      lgt ppqn (+ lgt first-lgt)))

          (t (error "display-length-of-midi-list. Unknown mode. Should either be absTime or deltaTime (string)")))))


(defun length-of-midi-list ()
  "Return the length of the selected midi list (an integer).
Works in both (clean and exclusive) absTime and deltaTime modes.
Does not include the first deltaTime if applied on a selection of deltaTime midi messages."
  (let* ((selected-txt (buffer-substring-no-properties (min (point) (mark)) (max (point) (mark))))
         (selected-lst (car (read-from-string (concat "(" selected-txt ")"))))
         (mode (guess-processing-mode  (min (point) (mark)) (max (point) (mark)))) ; a string: "absTime" or "deltaTime"   
        )
    (cond ((equal mode "absTime")
            (let ((time-of-first (string-to-number (ast-list-attribute (car selected-lst) 'absTime)))
                  (time-of-last  (string-to-number (ast-list-attribute (car (last selected-lst)) 'absTime))))
              (- time-of-last time-of-first))              
          )
          ((equal mode "deltaTime")
              (if (> (length selected-lst) 1)
                  (accumulate-right (function +) 0 (mapcar (function (lambda (ast-lst) (string-to-number (ast-list-attribute ast-lst 'deltaTime)))) (cdr selected-lst)))
                  0))
          (t (error "length-of-midi-list: Unknown mode. Should be either absTime or deltaTime")))))


(defun length-of-first-message-of-current-buffer ()
  "Return the length of the first event of the selected midi list (an integer).
Works in both (clean and exclusive) absTime and deltaTime modes."
  (let* ((selected-txt (buffer-substring-no-properties (min (point) (mark)) (max (point) (mark))))
         (mode (guess-processing-mode  (min (point) (mark)) (max (point) (mark)))) ; a string: "absTime" or "deltaTime"
        )
    (cond ((equal mode "absTime")
             (string-to-number (ast-list-attribute (car (read-from-string (concat selected-txt))) 'absTime)))
          ((equal mode "deltaTime")
             (string-to-number (ast-list-attribute (car (read-from-string (concat selected-txt))) 'deltaTime)))
          (t (error " length-of-first-message-of-current-buffer: Unknown mode. Should be either absTime or deltaTime")))))

(defun ast-list-attribute (ast-lst attribute-name) 
  "Primitive version of the Scheme function ast-attribute. 
ast-lst is the emacs list form of midi message.
Return the attribute value (a string) of attribute-name (a symbol).
Return nil if attribute-name is not found in the ast-lst."
  (if (not (null ast-lst))
      (ast-list-attribute-1 (cdr ast-lst) attribute-name)))

(defun ast-list-attribute-1 (ast-lst attribute-name)
  (cond ((null ast-lst) nil)                                                                              ; empty list
        ((and (not (null (cdr ast-lst))) (symbolp (car ast-lst)) (eq? (car ast-lst) attribute-name))      ; 'an "val"
            (cadr ast-lst))
        ((and (not (null (cdr ast-lst))) (listp (car ast-lst))                                            ; (quote an) "val2
              (= (length (car ast-lst)) 2) (eq (caar ast-lst) 'quote) (eq (cadr (car ast-lst)) attribute-name))
              (cadr ast-lst))
        (t                                                                                                ; everything else...
              (ast-list-attribute-1 (cdr ast-lst) attribute-name))))
 
; ---------------------------------------------------------------------------------------------------------------
; Bar drawing.

; Internal global variable of bar drawing:
(defvar previous-bar-number nil) 

(defun draw-bars-of-midi-track-liberal ()
  "Draw bar comments of the current MidiTrack.
   The point is supposed to be located at the start parenthesis of a MidiTrack, or inside a MidiTrack form (depth 1 or 2, not inside a string)."
  (interactive)
  (save-excursion
    (let ((p0 (point)))
      (if (looking-at "(MidiTrack")
          (progn (draw-bars-of-midi-track) "Done")
        (progn
          (goto-char p0)
          (backward-up-list 1)
          (if (looking-at "(MidiTrack")
              (progn (draw-bars-of-midi-track)
                     (message "Done")
                     "Done")
            (progn
              (goto-char p0)
              (backward-up-list 2)
              (if (looking-at "(MidiTrack")
                  (progn (draw-bars-of-midi-track)
                          (message "Done")
                          "Done"
                  )
                  (message "You should locate point at or inside a MidiTrack  (outside strings, max depth 2). Nothing done.")))))))))

; Precondition: The start parenthesis of a MidiTrack form is selected
(defun draw-bars-of-midi-track ()
  (let ((p0 (point)))
    (forward-sexp 1)
    (let ((pn (point)))
      (goto-char pn)
      (let ((next-bar-number (find-previous-bar-number)))
        (while (and next-bar-number (> (point) p0))
          (if (not (eq next-bar-number previous-bar-number))
              (progn
                (beginning-of-line 1) (next-line 1)
                (if (null previous-bar-number) (setq previous-bar-number (+ next-bar-number 1)))  ; only first time - at the end of the track
                (draw-bar-line previous-bar-number) 
                (beginning-of-line 1) (previous-line 1)
;                (end-of-line 1)
                (setq previous-bar-number next-bar-number)
                (setq next-bar-number (find-previous-bar-number)))
            (setq next-bar-number (find-previous-bar-number))))))))
       
(defun find-next-bar-number ()
  "Starting from point, find the next bar number from an info attribute. Moves point to a position after the info attribute name.
   Returns nil if no next bar number can be found."
  (if (and (search-forward "'info" nil t) (search-forward "\"" nil t))
      (let ((p0 (point))
            (pn (search-forward ":" nil t)))
        (string-to-number (buffer-substring-no-properties p0 (- pn 1))))
      nil))

(defun find-previous-bar-number ()
  "Starting from point, find the previous bar number from an info attribute. Moves point to a position before the info attribute name.
   Returns nil if no next bar number can be found."
  (if (and (search-backward "'info" nil t) (search-forward "\"" nil t))
      (let ((p0 (point))
            (pn (search-forward ":" nil t)))
        (search-backward "'info" nil t)
        (string-to-number (buffer-substring-no-properties p0 (- pn 1))))
      nil))

(defun find-max-bar-number ()
  (save-excursion
    (beginning-of-line)  ; to exist strings 
    (backward-up-list 1) ; looking at MidiTrack
    (forward-sexp 1)     ; at end of MidiTrack
    (find-previous-bar-number)
  )
)
    

; precondition: point is at the beginning of the first line of the new bar
(defun draw-bar-line (bar-number)
  (let* ((offset (cond ((< bar-number 10) 2)
                       ((< bar-number 100) 1)
                       (t 0)))
         (bar-line-length (+ score-start-column offset 18))
        ) 
    (insert (concat "; "
                    (number-to-string bar-number) " "
                    (make-string bar-line-length ?-)
                    CR))))



(defun remove-bars-of-midi-track-liberal ()
  (interactive)
  (save-excursion
    (let ((p0 (point)))
      (if (looking-at "(MidiTrack")
          (progn (remove-bars-from-midi-track) "Done")
        (progn
          (goto-char p0)
          (backward-up-list 1)
          (if (looking-at "(MidiTrack")
              (progn (remove-bars-from-midi-track) "Done")
            (progn
              (goto-char p0)
              (backward-up-list 2)
              (if (looking-at "(MidiTrack")
                  (progn (remove-bars-from-midi-track) "Done")
                  (message "You should locate point at or inside a MidiTrack  (outside strings, max depth 2). Nothing done.")))))))))

(defun remove-bars-from-midi-track ()
  (while (search-forward "; " nil t)
    (if (is-point-inside-bar-comment-p)
        (remove-current-bar-comment))))

(defvar bar-comment-string (make-string score-start-column ?-))

(defun is-point-inside-bar-comment-p ()
  (save-excursion
    (beginning-of-line)
    (let ((p0 (point)))
      (end-of-line)
      (let ((pn (point)))
        (goto-char p0)
        (if (search-forward "; " pn t)
            (if (looking-at-chars (point) (list ?0 ?1 ?2 ?3 ?4 ?5 ?6 ?7 ?8 ?9))
                (if (search-forward bar-comment-string pn t)
                    t
                  nil)
              nil)
          nil)))))

(defun remove-current-bar-comment ()
  (beginning-of-line)
  (kill-line 1))


; ---------------------------------------------------------------------------------------------------------------

(defun goto-bar (bar-number)
  (interactive "nGoto bar number: ")
  (let ((p0 (point)))
    (if (looking-at-track-like-form)
        (goto-bar-1 bar-number p0) 
        (progn
          (goto-char p0)
          (backward-up-list 1)
          (if (looking-at-track-like-form)
              (goto-bar-1 bar-number p0)
              (progn
                (goto-char p0)
                (backward-up-list 2)
                (if (looking-at-track-like-form)
                    (goto-bar-1 bar-number p0)
                    (message "Cannot goto a bar number from this context."))))))))

(defvar track-like-form-names 
  (list "MidiTrack" "MainA" "MainB" "MainC" "MainD" "FillInAA" "FillInBB" "FillInCC" "FillInDD"  "FillInBA" "FillInAB" 
        "IntroA" "IntroB" "IntroC" 
        "EndingA" "EndingB" "EndingC" ))

(defun looking-at-track-like-form ()
  (save-excursion
    (while (looking-at-chars (point) (list 40           ; start-parenthesis
                                           9 10 13 32   ; white space
                                     )) (forward-char 1))
    (member (name-under-point) track-like-form-names )))
        

(defun goto-bar-1 (bar-number p0)
  (let ((bn (find-next-bar-number)))
    (while (and bn (not (eq bn bar-number)))
      (setq bn (find-next-bar-number)))
    (if (eq bar-number bn)
        (beginning-of-line 1)
        (progn (goto-char p0)
               (message (concat "Cannot goto bar " (number-to-string bar-number)))))))

; ---------------------------------------------------------------------------------------------------------------
; Capturing stuff

(defvar capture-dir "c:/users/kurt/Media/midi/Cantabile/")
(defvar capturing-program "Cantabile")

(defun cantabile-help ()
  (interactive)
  (let ((help-buffer (if (get-buffer "*Help*") (get-buffer "*Help*") (generate-new-buffer "*Help*"))))
    (delete-other-windows)
    (split-window-vertically)
    (set-buffer help-buffer)
    (show-buffer (other-window 1) help-buffer)
    (toggle-read-only -1)
    (erase-buffer)
    (insert-file-contents (concat laml-dir "styles/xml-in-laml/midi/" "help/" "cantabile-help.txt"))
    (toggle-read-only 1)
    (set-buffer-modified-p nil)
    (other-window 1)))

(defun capture-midi-file()
  "Watch and wait for a new file in capture-dir. If or when it appears, parse it with midi-to-laml and show it in a buffer"
  (interactive)
  (let ((initial-file-list (directory-files capture-dir))
        (current-file-list (directory-files capture-dir))
        (i 0)
       )
    (while (= (length initial-file-list) (length current-file-list))
      (sleep-for 2) 
      (setq i (+ i 1))
      (message (make-string i ?.))
      (setq current-file-list (directory-files capture-dir)))

    (let ((new-files (list-difference current-file-list initial-file-list)))
       (if (= (length new-files) 1)    
           (progn
             (sync-midi-file-to-laml-file "deltaTime" "0" nil (concat capture-dir (car new-files)) (concat capture-dir (file-name-proper (car new-files)) "." "midl"))
             (find-file (concat capture-dir (file-name-proper (car new-files)) "." "midl"))
           )
           (error "Could not locate new midi file")))))
    

; To general
(defun string-search-forward (in-string search-string &optional from to)
  "Search for first occurrence of search-string in in-string within
the index interval from to. from is assumed to be smaller than to.
Return the index of the beginning of the first match, or nil
if no match is found. All indexes are zero bound.
The parameter from defaults to 0 and to defaults to (length in-string)."
  (let* ((from-1 (if from from 0))
         (to-1 (if to to (length in-string)))
         (temp-buf-name "*search-buffer*")
         (temp-buf (if (get-buffer temp-buf-name)
                       (get-buffer temp-buf-name)
                       (generate-new-buffer temp-buf-name)))
         (cur-buffer (current-buffer)))
    (set-buffer temp-buf)
    (erase-buffer)
    (insert in-string)
    (goto-char (+ from-1 1))
    (let ((res (search-forward search-string (+ to-1 1) t)))
      (kill-buffer temp-buf)
      (set-buffer cur-buffer)
      (if res (- res (+ (length search-string) 1)) nil)
      )))

; To general
(defun string-search-backward (in-string search-string &optional from to)
  "Search for last occurrence of search-string in in-string within
the index interval from to. from is assumed to larger than to.
Return the index of the beginning of the last match, or nil
if no match is found. All indexes are zero bound.
The parameter from defaults to 0 and to defaults to (length in-string)."
  (let* ((from-1 (if from from (length in-string)))
         (to-1 (if to to 0))
         (temp-buf-name "*search-buffer*")
         (temp-buf (if (get-buffer temp-buf-name)
                       (get-buffer temp-buf-name)
                       (generate-new-buffer temp-buf-name)))
         (cur-buffer (current-buffer)))
    (set-buffer temp-buf)
    (erase-buffer)
    (insert in-string)
    (goto-char (+ from-1 1))
    (let ((res (search-backward search-string (+ to-1 1) t)))
      (kill-buffer temp-buf)
      (set-buffer cur-buffer)
      (if res (- res 1) nil)
      )))


(defun midl-mode-status-string ()
;  (if (is-midl-stack-inconsistent) (recover-midl-source-file-stack))
  (concat
     ; Midl stack:
     (if (and (null midl-source-file-stack) (null midl-source-file-ghost-stack))
         (concat "Stack: Only this entry")
       (concat "Stack: " (midl-stack-image midl-source-file-stack midl-source-file-ghost-stack))) "  "

     ; State:
     "Attribute: " (pad-string-to-length 8 attribute-to-be-modified) " - "
     (if play-after-navigate
         "Navigate and play"
         "Navigate only    ") " - "
     "Channels: " (channel-info-string channels-to-play)
  )
)

(defun midl-stack-image (midl-source-file-stack midl-source-file-ghost-stack)
  (concat
     (make-string (length midl-source-file-stack) ?-)
     "*"
     (make-string (length midl-source-file-ghost-stack) ?.))) 
  

(defun pad-string-to-length (lgt str &optional justification)
 "Pad the string strt to a total length.
justification is either left or right (a symbol), and it defaults to 'left"
 (let ((just (if (not justification) 'left justification)))
  (let ((str-lgt (length str)))
    (cond ((eq just 'left) (concat str (make-string (- lgt str-lgt) 32)))
          ((eq just 'right) (concat (make-string (- lgt str-lgt) 32) str))
          (t (error "pad-string-to-length: Unknown justification" just))))))

(defun show-midl-mode-status-string ()
; (setq global-mode-string (midl-mode-status-string))
  (setq local-midl-mode-info (midl-mode-status-string))
  (force-mode-line-update))
    

(defun help-from-text-file (file)
  "Display help from file in help directory of midi stuff."
  (let ((help-buffer (if (get-buffer "*Help*") (get-buffer "*Help*") (generate-new-buffer "*Help*"))))
    (show-buffer (other-window 1) help-buffer)
    (set-buffer help-buffer)
    (toggle-read-only -1)
    (erase-buffer)
    (insert-file (concat laml-dir "styles/xml-in-laml/midi/help/" file))
    (toggle-read-only 1)
    (set-buffer-modified-p nil)
    (other-window 1)))

(defun open-midi-html-documentation (doc-file)
  (start-process "Windows Application" nil (concat laml-dir "bin/call.bat") (concat "c:/Programs/laml/styles/xml-in-laml/midi/man/" doc-file)))

; ---------------------------------------------------------------------------------------------------------------
; Support of marked sections (by use of meta markers).
; Markers are hosted as textual contetns in type 6 meta events. Format: "M-i<space>free-text" where i is an integer number.

(defun locate-initial-point-of-marked-section (marker-number)
 "Return the start point of the section that starts with a marker numbered with marker-number.
Return nil if the given marker-number cannot be found.
The marker number is an integer, or a string which represents an integer.
This function works relative to the current track."
   (find-marker-position marker-number))

(defun locate-final-point-of-marked-section (marker-number)
 "Return the end point of the section that starts with a marker numbered with marker-number.
The marker number is an integer, or a string which represents an integer.
This function works relative to the current track."
 (save-excursion
   (let ((p-ini (find-marker-position marker-number)))
     (if p-ini
         (progn
           (goto-char p-ini) (end-of-line 1)
           (find-next-marker-position)
         )
         nil))))


(defun find-marker-position (n)
 "Return the position of the meta form with marker n (an integer or string which represents the integer). 
Operates within the currently, implicitly selected track"
 (save-excursion
  (beginning-of-line 1) (backward-up-list 1)  ; now at beginning of the current MidiTrack
  (forward-sexp 1)
  (let ((limit (point)))
     (backward-sexp 1)
     (let* ((n-str (if (integerp n) (int-to-string n) n)))
       (find-marker-string-start-at (point) n-str limit)))))


(defun find-next-marker-position (&optional track-limit)   ; use of track-limit makes the function much more efficient
  "Relative the current point, return the position of the meta form with the next occuring mark in this track."
  (save-excursion
   (if track-limit  
       (find-marker-string-start-at (point) nil track-limit)
       (let ((cur-point (point)))
         (beginning-of-line 1) (backward-up-list 1) ; now at beginning of the current MidiTrack
         (forward-sexp 1)
         (let ((limit (point)))
           (goto-char cur-point)
           (find-marker-string-start-at cur-point nil limit))))))

(defun find-previous-marker-position (&optional track-limit) ; use of track-limit makes the function much more efficient
  "Relative the current point, return the position of the meta form with the previous occuring mark in this track."
  (save-excursion
   (if track-limit
       (find-previous-marker-string-start-at (point) track-limit)
       (let ((cur-point (point)))
         (beginning-of-line 1) (backward-up-list 1) ; now at beginning of the current MidiTrack
         (let ((limit (point)))
           (goto-char cur-point)
           (find-previous-marker-string-start-at cur-point limit))))))

(defun find-previous-marker-string-start-at (p limit)
  "Find the position of previous meta marker (of arbitrary number) between position p and limit. (Thus, search backward from p towards limit).
Return the found position.
Return nil if the marker cannot be located."
  (goto-char p)
  (let ((pos (search-backward "M-" limit t)))
    (if pos
        (progn
          (backward-char 1) (backward-up-list 1)
          (if (looking-at "(Meta")
              (point)
              (find-previous-marker-string-start-at (- pos 1) limit)))
        nil)))

(defun find-marker-string-start-at (p n-str limit)
  "Find the marker string n-str, such as \"5\", between position p and limit. Return the found position.
If n-str is nil, find the first marker following p.
Return nil if the marker cannot be located."
  (goto-char p)
  (let ((pos (search-forward (concat "M-" (if (null n-str) "" (concat n-str " "))) limit t)))
    (if pos
        (progn
          (backward-char (+ 4 (length n-str)))
          (backward-up-list 1)
          (if (looking-at "(Meta")
              (point)
              (find-marker-string-start-at (+ pos 1) n-str limit)))
        nil)))

(defun start-position-of-current-track ()
 "Return the position just before the first event of the current track."
 (save-excursion
   (beginning-of-line 1) 
   (backward-up-list 1) ; now at beginning of the current MidiTrack   
   (if (looking-at "(MidiTrack")
       (progn
         (forward-char 1)
         (forward-sexp 2)
         (backward-sexp 1)
         (point))
       nil)))


(defun end-position-of-current-track ()
  "Return the position just after the last event of the current track."
 (save-excursion
   (beginning-of-line 1) 
   (backward-up-list 1) ; now at beginning of the current MidiTrack   
   (if (looking-at "(MidiTrack")
       (progn
         (forward-sexp 1)
         (backward-char 1)
         (point))
       nil)))

(defun process-current-marked-region-and-play ()
  (interactive)
  (let* ((start-p (find-previous-marker-position))
         (end-p   (find-next-marker-position))
         (start-p-1 (if start-p start-p (start-position-of-current-track))) 
         (end-p-1 (if end-p end-p (end-position-of-current-track))) 
         (current-marker (marker-number-of-current-section))
        )
    (if (and start-p-1 end-p-1)  ; expected always to be true
        (progn
          (setq selection-begin start-p-1) (setq selection-end end-p-1) 
          (process-laml-file-and-present-from-to (channel-list-as-string channels-to-play) start-p-1 end-p-1 'play)
          (message "Playing section starting with the marker M-%s" current-marker)
        )
        (message "Cannot process and play marked region."))))

(defun goto-next-marked-section-and-process-it-and-play ()
  (interactive)
  (let ((start-p (find-next-marker-position)))
    (if start-p
        (progn
          (goto-char start-p)
          (forward-midi-message 1)
          (if play-after-navigate
              (let* ((current-marker (marker-number-of-current-section))
                     (end-p (find-next-marker-position))
                     (end-p-1 (if end-p end-p (end-position-of-current-track))) )
                (if end-p-1
                    (progn
                      (setq selection-begin start-p) (setq selection-end end-p-1) 
                      (message "Playing section starting with the marker M-%s" current-marker)
                      (process-laml-file-and-present-from-to (channel-list-as-string channels-to-play) start-p end-p-1 'play)
                      )
                  (message "Cannot play marked section")))
              (message "Now at section starting with the marker M-%s" (marker-number-of-current-section))))               
         (message "At End."))))


(defun goto-previous-marked-section-and-process-it-and-play ()
  (interactive)
  (let ((end-p (find-previous-marker-position)))
    (if end-p
        (progn
          (goto-char end-p)
          (let* ((current-marker (marker-number-of-current-section))
                 (start-p (find-previous-marker-position))
                 (start-p-1 (if start-p start-p (start-position-of-current-track))) )
             (if start-p-1
                 (progn
                   (goto-char start-p-1)
                   (forward-midi-message 1)
                   (if play-after-navigate
                       (progn
                         (setq selection-begin start-p-1) (setq selection-end end-p) 
                         (message "Playing section starting with the marker M-%s" (if current-marker current-marker "Begin"))
                         (process-laml-file-and-present-from-to (channel-list-as-string channels-to-play) start-p-1 end-p 'play))
                       (if current-marker
                          (message "Now at section starting with the marker M-%s" current-marker)
                          (message "Now at beginning"))))
                 (message "Cannot move to marked section"))))
         (message "At Beginning."))))


(defun extract-marker-text ()
  "Assume that the point is located at the start parenthesis of a Meta marker midi event with some M-i marking.
Extract full the marker text. A string. Return nil if not possible."
 (save-excursion
  (forward-sexp 1)
  (let ((limit (point)))
    (backward-sexp 1)
    (let ((start-p (search-forward "\"M-" limit t)))
      (if start-p
          (progn
            (goto-char (+ start-p 1))
            (let ((end-p (search-forward "\"" limit t)))
               (if end-p
                   (buffer-substring-no-properties (- start-p 2) (- end-p 1))
                   nil)))
          nil)))))

(defun marker-number-of-current-section ()
  "Return the marker number as a string (the marker number i of the previous meta marker M-i) of the current section.
A section goes from one meta marker to the next. Returns a string."
 (save-excursion
  (let ((p (find-previous-marker-position)))
    (if p
        (progn
           (goto-char p)
           (let ((res (search-forward "\"M-" nil t)))
             (if res
                 (let ((marker-name (name-under-point)))
                   (substring marker-name 2 (length marker-name)))
               nil)))
        nil))))

(defun collect-marker-list-of-track ()
 "Given the implicitly selected track, collect and retun a list of all marker texts in this track."
 (save-excursion
   (beginning-of-line 1) (backward-up-list 1) ; now at beginning of the current MidiTrack
   (forward-sexp 1)
   (let ((limit (point))
         (pos nil)
         (res nil)
         (i 0)
        )
     (forward-sexp -1)
     (while (setq pos (find-next-marker-position limit))
        (goto-char pos)
        (setq res (cons (extract-marker-text) res))
        (setq i (+ i 1))
        (message "Collecting marker %d" i)
        (forward-sexp 1)  ; progress
     )
     (reverse res))))

; ---------------------------------------------------------------------------------------------------------------
; Marker browser

(defun make-marker-region-browser ()
  (interactive)
  (let ((marker-text-list (filter (function (lambda (x) (not (null x)))) (collect-marker-list-of-track)))) 
    (delete-other-windows)
    (split-window-vertically)
    (let ((browser-buf (get-buffer-create "marker-browser"))
         )
      (set-buffer browser-buf)
      (toggle-read-only -1)
      (erase-buffer)
      (show-buffer (selected-window) browser-buf)
      (mapcar (function (lambda (mt) (insert mt) (insert CR))) marker-text-list)
      (goto-char (point-min))
      (marker-browser-mode)
      (toggle-read-only 1)
    )))
      
(define-derived-mode 
	marker-browser-mode fundamental-mode "MarkerBrower" 
  "..."

)  

(defvar marker-browser-mode-map (make-sparse-keymap "MarkerBrowser"))
(define-key marker-browser-mode-map "f" 'mb-goto-meta-marker)
(define-key marker-browser-mode-map "g" 'mb-goto-meta-marker)  ; alias of f
(define-key marker-browser-mode-map "p" 'mb-play-marked-region)
(define-key marker-browser-mode-map "c" 'mb-add-comment-to-marker)
(define-key marker-browser-mode-map  "l" 'mb-display-length-of-marked-region)
(define-key marker-browser-mode-map  "e" 'mb-enclose-marked-region)

(laml-define-key marker-browser-mode-map [menu-bar marker-browser]
    (cons "Marker Browser" (make-sparse-keymap "Marker Browser")))

(laml-define-key  marker-browser-mode-map [menu-bar marker-browser help-on-markers-from-browser]
      '("Help on Markers" . (lambda () (interactive) (help-from-text-file "marker-channel.txt"))))

(laml-define-key marker-browser-mode-map [menu-bar marker-browser marker-browser-sep-1]
   '("----"))

(laml-define-key  marker-browser-mode-map [menu-bar marker-browser mb-display-length-of-marked-region]
      '("Display length of region" . mb-display-length-of-marked-region))

(laml-define-key  marker-browser-mode-map [menu-bar marker-browser mb-enclose-marked-region]
      '("Enclose section in midi-region" . mb-enclose-marked-region))

(laml-define-key  marker-browser-mode-map [menu-bar marker-browser mb-add-comment-to-marker]
      '("Add comment to marker" . mb-add-comment-to-marker))

(laml-define-key  marker-browser-mode-map [menu-bar marker-browser mb-goto-meta-marker]
      '("Goto marker" . mb-goto-meta-marker))

(laml-define-key  marker-browser-mode-map [menu-bar marker-browser mb-play-marked-region]
      '("Play marked region" . mb-play-marked-region))








(defun marker-number-of-browser-line ()
  "Return the number - as a string - of the currently selected marker line in a marker browser."
  (save-excursion
   (beginning-of-line)
   (let ((start-p (search-forward "M-" nil t)))
    (if start-p
        (progn
          (forward-word 1)
          (let ((end-p (point)))
             (buffer-substring-no-properties start-p end-p)))
        nil  ; not expectable
    ))))
        

(defun mb-goto-meta-marker ()
  (interactive)
  (let ((marker-number-text (marker-number-of-browser-line)))
    (other-window 1)
    (let ((pos (find-marker-position marker-number-text)))
      (if pos
          (goto-char pos)
          (message "Cannot find marker"))
      (other-window 1))))


(defun mb-play-marked-region ()
  (interactive)
  (save-excursion
    (if (and transient-mark-mode mark-active) 

        ; Highlighed selection in browser buffer:
        (let ((mb-start-p (min (point) (mark)))
              (mb-end-p (max (point) (mark))))
          (goto-char mb-start-p)
          (let ((from-marker-number-text (marker-number-of-browser-line)))
            (goto-char mb-end-p)
            (let ((to-marker-number-text (marker-number-of-browser-line)))
              (other-window 1)
              (let* ((start-p (locate-initial-point-of-marked-section from-marker-number-text))
                     (end-p   (locate-final-point-of-marked-section to-marker-number-text))
                     (start-p-1 (if start-p start-p (start-position-of-current-track))) 
                     (end-p-1 (if end-p end-p (end-position-of-current-track))) 
                     (current-marker (marker-number-of-current-section))
                     )
                (if (and start-p-1 end-p-1) ; expected always to be true
                    (progn
                      (setq selection-begin start-p-1) (setq selection-end end-p-1) 
                      (process-laml-file-and-present-from-to (channel-list-as-string channels-to-play) start-p-1 end-p-1 'play)
                      )
                  (message "Cannot process and play marked region.")))
               (other-window 1))))

        ; Single browser line - implicit selection
        (let ((marker-number-text (marker-number-of-browser-line)))
          (if marker-number-text
              (progn
                (other-window 1)
                (let* ((start-p (locate-initial-point-of-marked-section marker-number-text))
                       (end-p   (locate-final-point-of-marked-section marker-number-text))
                       (start-p-1 (if start-p start-p (start-position-of-current-track))) 
                       (end-p-1 (if end-p end-p (end-position-of-current-track))) 
                       (current-marker (marker-number-of-current-section))
                       )
                  (if (and start-p-1 end-p-1) ; expected always to be true
                      (progn
                        (setq selection-begin start-p-1) (setq selection-end end-p-1) 
                        (process-laml-file-and-present-from-to (channel-list-as-string channels-to-play) start-p-1 end-p-1 'play)
                        )
                    (message "Cannot process and play marked region.")))
                (other-window 1))
            (message "Cannot??!!"))))))

(defun mb-add-comment-to-marker (comment)
  (interactive "sComment of this marker: ")

  (let ((marker-number-text (marker-number-of-browser-line)))
    (other-window 1)
    (let ((pos (find-marker-position marker-number-text)))
      (if pos
          (progn
            (goto-char pos)
            (forward-sexp 1)
            (let ((limit (point)))
              (backward-sexp 1)
              (let ((res (search-forward "\"M-" limit t)))
                (if res
                    (progn
                      (search-forward "\"" limit t)
                      (backward-char 1)
                      (insert " ") (insert comment))
                  (message "Cannot??!! (1)")))))
           (message "Cannot??!! (2)")))
      (other-window 1)

      ; now back in browser buffer
      (end-of-line 1)
      (toggle-read-only)
      (insert " ") (insert comment)
      (toggle-read-only)
      (beginning-of-line 1)
  ))

(defun mb-display-length-of-marked-region ()
  (interactive)
  (if (and transient-mark-mode mark-active) 

                                        ; Highlighed selection in browser buffer:
      (let ((mb-start-p (min (point) (mark)))
            (mb-end-p (max (point) (mark))))
        (goto-char mb-start-p)
        (let ((from-marker-number-text (marker-number-of-browser-line)))
          (goto-char mb-end-p)
          (let ((to-marker-number-text (marker-number-of-browser-line)))
            (if (and from-marker-number-text to-marker-number-text)
                (progn
                  (other-window 1)
                  (let* ((initial-p0 (locate-initial-point-of-marked-section from-marker-number-text))
                         (initial-p (if initial-p0 initial-p0 (start-position-of-current-track)))
                         (final-p0 (locate-final-point-of-marked-section to-marker-number-text))
                         (final-p (if final-p0 final-p0 (end-position-of-current-track)))) 
                    (set-mark initial-p)
                    (goto-char final-p)
                    (display-length-of-midi-list)
                    (other-window 1)))
              (message "Cannot")))))

    (let ((marker-number-text (marker-number-of-browser-line)))
      (if marker-number-text
          (save-excursion
            (other-window 1)
            (let* ((initial-p0 (locate-initial-point-of-marked-section marker-number-text))
                   (initial-p (if initial-p0 initial-p0 (start-position-of-current-track)))
                   (final-p0 (locate-final-point-of-marked-section marker-number-text))
                   (final-p (if final-p0 final-p0 (end-position-of-current-track)))) 
              (set-mark initial-p)
              (goto-char final-p)
              (display-length-of-midi-list)
              (other-window 1)))
        (message "Cannot")))))

(defun mb-enclose-marked-region ()
  "Enclose the marked region in a midi-region"
  (interactive)
  (if (and transient-mark-mode mark-active) 

                                        ; Highlighed selection in browser buffer:
      (let ((mb-start-p (min (point) (mark)))
            (mb-end-p (max (point) (mark))))
        (goto-char mb-start-p)
        (let ((from-marker-number-text (marker-number-of-browser-line)))
          (goto-char mb-end-p)
          (let ((to-marker-number-text (marker-number-of-browser-line)))
            (if (and from-marker-number-text to-marker-number-text)
                (progn
                  (other-window 1)
                  (let* ((initial-p0 (locate-initial-point-of-marked-section from-marker-number-text))
                         (initial-p (if initial-p0 initial-p0 (start-position-of-current-track)))
                         (final-p0 (locate-final-point-of-marked-section to-marker-number-text))
                         (final-p (if final-p0 final-p0 (end-position-of-current-track))))
                     (goto-char final-p) (backward-char 1)
                     (while (looking-at-white-space (point)) (backward-char 1))
                     (insert ")")

                     (goto-char initial-p)
                     (let ((indentation (current-indentation)))
                       (forward-sexp 1)     
                       (insert CR)
                       (insert (make-string (- indentation 1) 32)) (insert "(midi-region "))
                     (other-window 1)))
              (message "Cannot")))))

    (let ((marker-number-text (marker-number-of-browser-line)))
      (if marker-number-text
          (save-excursion
            (other-window 1)
            (let* ((initial-p0 (locate-initial-point-of-marked-section marker-number-text))
                   (initial-p (if initial-p0 initial-p0 (start-position-of-current-track)))
                   (final-p0 (locate-final-point-of-marked-section marker-number-text))
                   (final-p (if final-p0 final-p0 (end-position-of-current-track))))
              (goto-char final-p) (backward-char 1)
              (while (looking-at-white-space (point)) (backward-char 1))
              (insert ")")

              (goto-char initial-p)
              (let ((indentation (current-indentation)))
                (forward-sexp 1)     
                (insert CR)
                (insert (make-string (- indentation 1) 32)) (insert "(midi-region "))

             (other-window 1)))
        (message "Cannot")))))                


; ---------------------------------------------------------------------------------------------------------------
; Midl file comments:

(defun comment-of-midl-file ()
  "Return the current midl file comment"
  (save-excursion
     (goto-char (point-min))
     (if (looking-at ";;;; ")
         (progn
           (forward-char 5)
           (let ((p (point)))
             (end-of-line 1)
             (buffer-substring-no-properties p (point))))
         "")))

(defun set-comment-of-midl-file (comment-str)
  "Arrange that the first line of the current midl file becomes a four semicolon Scheme comment with comment-str as contents.
comment-str must be without CRs"
  (save-excursion
    (goto-char (point-min))
    (cond ((looking-at ";;;;")  ; killing existing fours semicolon comment
            (kill-line 1)
            (insert (concat ";;;; " comment-str CR)))
          (t (insert (concat ";;;; " comment-str CR))))))


(defun extract-registered-point ()
  "Extract the registered point - at the very end of the buffer"
  (save-excursion
    (goto-char (point-max))
    (let ((limit (point)))
       (beginning-of-line 1)
       (let ((registered-point (search-forward ";;; " limit t)))
         (if registered-point
             (string-to-int (name-under-point))
             nil)))))


(defun register-current-point (p)
  "Register point p at the very end of the buffer"
  (save-excursion
    (goto-char (point-max))
    (beginning-of-line 1)
    (if (looking-at ";;; ")
        (progn
           (beginning-of-line 1)
           (kill-line 1)))
    (insert (concat ";;; " (int-to-string p)))))

; ---------------------------------------------------------------------------------------------------------------

(defun midi-load ()
  (interactive)
  (insert "(laml-style \"xml-in-laml/midi/midi\")") (comint-send-input)
  (insert "(load (string-append laml-dir \"styles/xml-in-laml/midi/midi-laml-processing-lib.scm\"))") (comint-send-input)
)